from datetime import datetime
from decimal import Decimal

from quiffen.core.investment import Investment


def test_create_investment():
    """Test creating an investment"""
    date = datetime(2022, 1, 1)
    investment = Investment(date=date)
    assert investment.action is None
    assert investment.security is None
    assert investment.price is None
    assert investment.memo is None
    assert investment.commission is None

    investment2 = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    assert investment2.action == "Buy"
    assert investment2.security == "Test Security"
    assert investment2.price == 100
    assert investment2.memo == "Test Memo"
    assert investment2.commission == 10


def test_eq_success():
    """Test that two investments are equal"""
    date = datetime(2022, 1, 1)
    investment = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    investment2 = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    assert investment == investment2


def test_eq_failure():
    """Test that two investments are not equal"""
    date = datetime(2022, 1, 1)
    investment = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    investment2 = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(20),
    )
    assert investment != investment2


def test_str_method():
    """Test the string representation of an investment"""
    date = datetime(2022, 1, 1)
    investment = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    assert str(investment) == (
        "Investment:\n\tDate: 2022-01-01 00:00:00\n\tAction: Buy\n\tSecurity: "
        "Test Security\n\tPrice: 100\n\tMemo: Test Memo\n\tCommission: 10"
    )


def test_to_qif():
    """Test the to_qif method"""
    date = datetime(2022, 1, 1)
    investment = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    assert investment.to_qif() == (
        "D2022-01-01\nNBuy\nYTest Security\nI100\nMTest Memo\nO10\n"
    )


def test_to_dict():
    """Test the to_dict method"""
    date = datetime(2022, 1, 1)
    investment = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    assert investment.to_dict() == {
        "date": date,
        "action": "Buy",
        "security": "Test Security",
        "price": Decimal("100"),
        "memo": "Test Memo",
        "commission": Decimal("10"),
        "amount": None,
        "cleared": None,
        "first_line": None,
        "line_number": None,
        "quantity": None,
        "to_account": None,
        "transfer_amount": None,
    }


def test_to_dict_with_ignore():
    """Test the to_dict method with ignore"""
    date = datetime(2022, 1, 1)
    investment = Investment(
        date=date,
        action="Buy",
        security="Test Security",
        price=Decimal(100),
        memo="Test Memo",
        commission=Decimal(10),
    )
    assert investment.to_dict(ignore=["action", "security"]) == {
        "date": date,
        "price": Decimal("100"),
        "memo": "Test Memo",
        "commission": Decimal("10"),
        "amount": None,
        "cleared": None,
        "first_line": None,
        "line_number": None,
        "quantity": None,
        "to_account": None,
        "transfer_amount": None,
    }


def test_from_list_no_custom_fields():
    """Test creating an investment from a list with no custom fields"""
    qif_list = [
        "D2022-01-01",
        "NBuy",
        "YTest Security",
        "I100",
        "MTest Memo",
        "O10",
    ]
    investment = Investment.from_list(qif_list)
    assert investment.date == datetime(2022, 1, 1)
    assert investment.action == "Buy"
    assert investment.security == "Test Security"
    assert investment.price == 100
    assert investment.memo == "Test Memo"
    assert investment.commission == 10


def test_from_list_with_custom_fields():
    """Test creating an investment from a list with custom fields"""
    setattr(Investment, "__CUSTOM_FIELDS", [])  # Reset custom fields
    qif_list = [
        "D2022-01-01",
        "NBuy",
        "YTest Security",
        "I100",
        "MTest Memo",
        "O10",
        "XCustom field 1",
        "Y9238479",
        "DT2022-01-01T00:00:00.000001",
    ]

    # Add custom fields
    Investment.add_custom_field(
        line_code="X",
        attr="custom_field_1",
        field_type=str,
    )
    Investment.add_custom_field(
        line_code="Y",
        attr="custom_field_2",
        field_type=Decimal,
    )
    Investment.add_custom_field(
        line_code="DT",  # Test multi-character line code
        attr="custom_field_3",
        field_type=datetime,
    )

    investment = Investment.from_list(qif_list)
    assert investment.date == datetime(2022, 1, 1)
    assert investment.action == "Buy"
    assert investment.security == "Test Security"
    assert investment.price == 100
    assert investment.memo == "Test Memo"
    assert investment.commission == 10
    assert investment.custom_field_1 == "Custom field 1"  # type: ignore
    assert investment.custom_field_2 == Decimal("9238479")  # type: ignore
    assert investment.custom_field_3 == datetime(2022, 1, 1, 0, 0, 0, 1)  # type: ignore
    setattr(Investment, "__CUSTOM_FIELDS", [])  # Reset custom fields



# pylint: disable=redefined-outer-name,protected-access
from datetime import datetime
from decimal import Decimal
from pathlib import Path

import pytest

from quiffen.core.category import (
    Category,
    CategoryType,
    add_categories_to_container,
    create_categories_from_hierarchy,
)
from quiffen.core.qif import Qif


@pytest.fixture
def tree():
    """Return a tree of categories.

    Parent (root)
    └─ Child1
       └─ Grandchild1
       └─ Grandchild2
    └─ Child2
    """
    parent = Category(name="Parent")
    child1 = Category(name="Child1")
    child2 = Category(name="Child2")
    grandchild1 = Category(name="Grandchild1")
    grandchild2 = Category(name="Grandchild2")
    parent.add_child(child1)
    parent.add_child(child2)
    child1.add_child(grandchild1)
    child1.add_child(grandchild2)
    return parent, child1, child2, grandchild1, grandchild2


def test_create_category():
    """Test creating a category"""
    category = Category(name="Test")
    assert category.name == "Test"
    assert category.parent is None
    assert not category.children
    assert category.hierarchy == "Test"
    assert category.category_type == "expense"


def test_eq_success():
    """Test equality"""
    category1 = Category(name="Test")
    category2 = Category(name="Test")
    assert category1 == category2

    parent1 = Category(name="Parent")
    parent2 = Category(name="Parent")
    child1 = Category(name="Child")
    child2 = child1.model_copy()
    parent1.add_child(child1)
    parent2.add_child(child2)
    assert parent1 == parent2


def test_eq_failure():
    """Test equality"""
    category1 = Category(name="Test1")
    category2 = Category(name="Test1", category_type=CategoryType.INCOME)
    assert category1 != category2

    parent1 = Category(name="Parent")
    parent2 = Category(name="Parent")
    child = Category(name="Child")
    parent1.add_child(child)
    assert parent1 != parent2


def test_str_method():
    """Test the string method"""
    category = Category(
        name="Test",
        category_type=CategoryType.INCOME,
        desc="Test",
        children=[Category(name="Child 1"), Category(name="Child 2")],
    )
    assert str(category) == (
        "Category:\n\tName: Test\n\tDesc: Test\n\t"
        "Category Type: income\n\tHierarchy: Test\n\tChildren: 2"
    )


def test_refresh_hierarchy():
    """Test refreshing the hierarchy of a category"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    parent.add_child(child)
    child.name = "New Name"
    parent._refresh_hierarchy()
    assert child.hierarchy == "Parent:New Name"


def test_to_dict(tree):
    """Test converting a category to a dictionary"""
    parent, child1, child2, grandchild1, grandchild2 = tree
    assert parent.to_dict() == {
        "name": "Parent",
        "category_type": "expense",
        "children": ["Child1", "Child2"],
        "hierarchy": "Parent",
        "budget_amount": None,
        "desc": None,
        "parent": None,
        "tax_related": None,
        "tax_schedule_info": None,
    }
    assert child1.to_dict() == {
        "name": "Child1",
        "category_type": "expense",
        "children": ["Grandchild1", "Grandchild2"],
        "hierarchy": "Parent:Child1",
        "parent": "Parent",
        "budget_amount": None,
        "desc": None,
        "tax_related": None,
        "tax_schedule_info": None,
    }
    assert child2.to_dict() == {
        "name": "Child2",
        "category_type": "expense",
        "children": [],
        "hierarchy": "Parent:Child2",
        "parent": "Parent",
        "budget_amount": None,
        "desc": None,
        "tax_related": None,
        "tax_schedule_info": None,
    }
    assert grandchild1.to_dict() == {
        "name": "Grandchild1",
        "category_type": "expense",
        "children": [],
        "hierarchy": "Parent:Child1:Grandchild1",
        "parent": "Child1",
        "budget_amount": None,
        "desc": None,
        "tax_related": None,
        "tax_schedule_info": None,
    }
    assert grandchild2.to_dict() == {
        "name": "Grandchild2",
        "category_type": "expense",
        "children": [],
        "hierarchy": "Parent:Child1:Grandchild2",
        "parent": "Child1",
        "budget_amount": None,
        "desc": None,
        "tax_related": None,
        "tax_schedule_info": None,
    }


def test_to_dict_with_ignore():
    """Test converting a category to a dictionary with ignored fields"""
    category = Category(name="Test")
    assert category.to_dict(ignore=["name"]) == {
        "category_type": "expense",
        "children": [],
        "hierarchy": "Test",
        "budget_amount": None,
        "desc": None,
        "parent": None,
        "tax_related": None,
        "tax_schedule_info": None,
    }


def test_set_parent():
    """Test setting a parent category"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    child.set_parent(parent)
    assert child.parent == parent
    assert parent.children == [child]
    assert child.hierarchy == "Parent:Child"


def test_set_parent_with_children():
    """Test setting a parent category that already has children"""
    parent = Category(name="Parent")
    child1 = Category(name="Child1")
    child2 = Category(name="Child2")
    child1.add_child(child2)
    child2.set_parent(parent)
    assert child1.parent is None
    assert child2.parent == parent
    assert parent.children == [child2]
    assert not child1.children
    assert not child2.children
    assert child1.hierarchy == "Child1"
    assert child2.hierarchy == "Parent:Child2"


def test_set_parent_to_self():
    """Test setting a category as its own parent raises a ValueError"""
    category = Category(name="Test")
    with pytest.raises(ValueError):
        category.set_parent(category)


def test_merge_single_level():
    """Test merging two categories, one level deep"""
    parent1 = Category(name="Parent")
    parent2 = Category(name="Parent")
    child1 = Category(name="Child1")
    child2 = Category(name="Child2")
    child3 = Category(name="Child3")
    parent1.add_child(child1)
    parent1.add_child(child2)
    parent2.add_child(child3)

    parent1.merge(parent2)

    assert sorted(parent1.children) == sorted([child1, child2, child3])
    assert child3.parent == parent1


def test_merge_multi_level():
    """Test merging two categories, multiple levels deep"""
    parent1 = Category(name="Parent")
    parent2 = Category(name="Parent")
    child1 = Category(name="Child1")
    child1_2 = Category(name="Child1")
    child2 = Category(name="Child2")
    grandchild1 = Category(name="Grandchild1")
    grandchild2 = Category(name="Grandchild2")
    grandchild3 = Category(name="Grandchild3")
    parent1.add_child(child1)
    parent1.add_child(child2)
    parent2.add_child(child1_2)
    parent2.add_child(child2)
    child1.add_child(grandchild1)
    child1_2.add_child(grandchild2)
    child2.add_child(grandchild3)

    parent1.merge(parent2)

    assert sorted(parent1.children) == sorted([child1, child2])
    assert sorted(child1.children) == sorted([grandchild1, grandchild2])
    assert sorted(child2.children) == sorted([grandchild3])
    assert grandchild1.parent == child1
    assert grandchild2.parent == child1
    assert grandchild3.parent == child2


def test_add_child():
    """Test adding a child category"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    parent.add_child(child)
    assert child.parent == parent
    assert parent.children == [child]
    assert child.hierarchy == "Parent:Child"


def test_add_multiple_children():
    """Test adding multiple children to a category"""
    parent = Category(name="Parent")
    child1 = Category(name="Child1")
    child2 = Category(name="Child2")
    parent.add_child(child1)
    parent.add_child(child2)
    assert child1.parent == parent
    assert child2.parent == parent
    assert parent.children == [child1, child2]
    assert child1.hierarchy == "Parent:Child1"
    assert child2.hierarchy == "Parent:Child2"


def test_add_child_with_parent():
    """Test adding a child category that already has a parent"""
    parent1 = Category(name="Parent1")
    parent2 = Category(name="Parent2")
    child = Category(name="Child")
    parent1.add_child(child)
    parent2.add_child(child)
    assert child.parent == parent2
    assert not parent1.children
    assert parent2.children == [child]
    assert child.hierarchy == "Parent2:Child"


def test_add_child_with_children():
    """Test adding a child category that already has children"""
    parent = Category(name="Parent")
    child1 = Category(name="Child1")
    child2 = Category(name="Child2")
    child3 = Category(name="Child3")
    child1.add_child(child2)
    parent.add_child(child1)
    parent.add_child(child3)
    assert child1.parent == parent
    assert child2.parent == child1
    assert child3.parent == parent
    assert parent.children == [child1, child3]
    assert child1.children == [child2]
    assert not child2.children
    assert child1.hierarchy == "Parent:Child1"
    assert child2.hierarchy == "Parent:Child1:Child2"
    assert child3.hierarchy == "Parent:Child3"


def test_add_child_multiple_times():
    """Test adding a child category multiple times"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    parent.add_child(child)
    parent.add_child(child)
    assert child.parent == parent
    assert parent.children == [child]
    assert child.hierarchy == "Parent:Child"


def test_add_child_multi_level():
    """Test adding a child to a category that has a parent"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    grandchild = Category(name="Grandchild")
    parent.add_child(child)
    child.add_child(grandchild)
    assert child.parent == parent
    assert grandchild.parent == child
    assert parent.children == [child]
    assert child.children == [grandchild]
    assert child.hierarchy == "Parent:Child"
    assert grandchild.hierarchy == "Parent:Child:Grandchild"


def test_add_self_as_child():
    """Test adding a category as its own child raises a ValueError"""
    category = Category(name="Test")
    with pytest.raises(ValueError):
        category.add_child(category)


def test_remove_child():
    """Test removing a child category"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    parent.add_child(child)
    parent.remove_child(child)
    assert child.parent is None
    assert not parent.children
    assert child.hierarchy == "Child"
    assert parent.hierarchy == "Parent"


def test_remove_child_with_children_without_keep():
    """Test removing a child category that has children"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    grandchild = Category(name="Grandchild")
    parent.add_child(child)
    child.add_child(grandchild)
    parent.remove_child(child)
    assert child.parent is None
    assert grandchild.parent == child
    assert not parent.children
    assert child.children == [grandchild]
    assert not grandchild.children
    assert child.hierarchy == "Child"
    assert grandchild.hierarchy == "Child:Grandchild"
    assert parent.hierarchy == "Parent"


def test_remove_child_with_children_with_keep():
    """Test removing a child category that has children"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    grandchild = Category(name="Grandchild")
    parent.add_child(child)
    child.add_child(grandchild)
    parent.remove_child(child, keep_children=True)
    assert child.parent is None
    assert grandchild.parent == parent
    assert parent.children == [grandchild]
    assert not child.children
    assert not grandchild.children
    assert child.hierarchy == "Child"
    assert grandchild.hierarchy == "Parent:Grandchild"
    assert parent.hierarchy == "Parent"


def test_remove_child_with_children_with_keep_multi_level():
    """Test removing a child category that has children"""
    parent = Category(name="Parent")
    child = Category(name="Child")
    grandchild = Category(name="Grandchild")
    great_grandchild = Category(name="GreatGrandchild")
    parent.add_child(child)
    child.add_child(grandchild)
    grandchild.add_child(great_grandchild)
    parent.remove_child(child, keep_children=True)
    assert child.parent is None
    assert grandchild.parent == parent
    assert great_grandchild.parent == grandchild
    assert parent.children == [grandchild]
    assert not child.children
    assert grandchild.children == [great_grandchild]
    assert not great_grandchild.children
    assert child.hierarchy == "Child"
    assert grandchild.hierarchy == "Parent:Grandchild"
    assert great_grandchild.hierarchy == "Parent:Grandchild:GreatGrandchild"
    assert parent.hierarchy == "Parent"


def test_render_tree(tree):
    """Test rendering a tree of categories"""
    parent, *_ = tree
    render = parent.render_tree()
    print(render)
    assert render == (
        "Parent (root)\n└─ Child1\n   └─ Grandchild1\n   └─ Grandchild2\n└─ Child2"
    )


def test_render_tree_no_tree():
    """Test rendering a tree of categories with no children"""
    root = Category(name="Parent")
    tree = root.render_tree()
    assert tree == "Parent"


def test_traverse_down(tree):
    """Test traversing down a tree"""
    parent, child1, child2, grandchild1, grandchild2 = tree

    parent_traversal = parent.traverse_down()
    assert len(parent_traversal) == 5
    for node in (parent, child1, grandchild1, grandchild2, child2):
        assert node in parent_traversal

    child1_traversal = child1.traverse_down()
    assert len(child1_traversal) == 3
    for node in (child1, grandchild1, grandchild2):
        assert node in child1_traversal

    assert child2.traverse_down() == [child2]
    assert grandchild1.traverse_down() == [grandchild1]
    assert grandchild2.traverse_down() == [grandchild2]


def test_traverse_up(tree):
    """Test traversing up a tree"""
    parent, child1, child2, grandchild1, grandchild2 = tree

    parent_traversal = parent.traverse_up()
    assert len(parent_traversal) == 1
    assert parent in parent_traversal

    child1_traversal = child1.traverse_up()
    assert len(child1_traversal) == 2
    for node in (child1, parent):
        assert node in child1_traversal

    child2_traversal = child2.traverse_up()
    assert len(child2_traversal) == 2
    for node in (child2, parent):
        assert node in child2_traversal

    grandchild1_traversal = grandchild1.traverse_up()
    assert len(grandchild1_traversal) == 3
    for node in (grandchild1, child1, parent):
        assert node in grandchild1_traversal

    grandchild2_traversal = grandchild2.traverse_up()
    assert len(grandchild2_traversal) == 3
    for node in (grandchild2, child1, parent):
        assert node in grandchild2_traversal


def test_find_child(tree):
    """Test finding a child"""
    parent, child1, child2, grandchild1, grandchild2 = tree
    assert parent.find_child("Child1") == child1
    assert parent.find_child("Child2") == child2
    assert parent.find_child("Grandchild1") == grandchild1
    assert parent.find_child("Grandchild2") == grandchild2
    assert child1.find_child("Grandchild1") == grandchild1
    assert child1.find_child("Grandchild2") == grandchild2

    assert child2.find_child("Grandchild1") is None
    assert child2.find_child("Grandchild2") is None
    assert grandchild1.find_child("Grandchild1") is grandchild1
    assert grandchild1.find_child("Grandchild2") is None
    assert grandchild2.find_child("Grandchild1") is None
    assert grandchild2.find_child("Grandchild2") is grandchild2


def test_to_qif():
    """Test converting a category to a QIF category"""
    root = Category(name="Root")
    child = Category(name="Child")
    root.add_child(child)
    assert root.to_qif() == (  # Should show both root and child
        "!Type:Cat\nNRoot\nE\n^\n!Type:Cat\nNRoot:Child\nE\n"
    )
    assert child.to_qif() == "!Type:Cat\nNRoot:Child\nE\n"


def test_to_qif_with_custom_fields():
    """Test converting a category to a QIF category"""
    setattr(Category, "__CUSTOM_FIELDS", [])  # Reset custom fields
    root = Category(name="Root")
    child = Category(name="Child")
    root.add_child(child)

    # Add custom fields
    Category.add_custom_field(
        line_code="X",
        attr="custom_field_1",
        field_type=str,
    )
    Category.add_custom_field(
        line_code="Y",
        attr="custom_field_2",
        field_type=Decimal,
    )
    Category.add_custom_field(
        line_code="DT",  # Test multi-character line code
        attr="custom_field_3",
        field_type=datetime,
    )

    root.custom_field_1 = "Custom field 1"  # type: ignore
    root.custom_field_2 = Decimal("9238479")  # type: ignore
    root.custom_field_3 = datetime(2022, 1, 1, 0, 0, 0, 1)  # type: ignore
    child.custom_field_1 = "Custom field 1"  # type: ignore
    child.custom_field_2 = Decimal("9238479")  # type: ignore
    child.custom_field_3 = datetime(2022, 1, 1, 0, 0, 0, 1)  # type: ignore

    assert (
        root.to_qif()
        == (  # Should show both root and child
            "!Type:Cat\n"
            "NRoot\n"
            "E\n"
            "DT2022-01-01 00:00:00.000001\n"
            "Y9238479\n"
            "XCustom field 1\n"
            "^\n"
            "!Type:Cat\n"
            "NRoot:Child\n"
            "E\n"
            "DT2022-01-01 00:00:00.000001\n"
            "Y9238479\n"
            "XCustom field 1\n"
        )
    )
    assert child.to_qif() == (
        "!Type:Cat\n"
        "NRoot:Child\n"
        "E\n"
        "DT2022-01-01 00:00:00.000001\n"
        "Y9238479\n"
        "XCustom field 1\n"
    )
    setattr(Category, "__CUSTOM_FIELDS", [])  # Reset custom fields


def test_from_list_no_custom_fields():
    """Test creating a category from a list of QIF strings"""
    qif_list = [
        "NParent",
        "DSome category description",
        "E",
        "T",
        "B123.45",
        "RTax schedule info",
    ]
    category = Category.from_list(qif_list)
    assert category.name == "Parent"
    assert category.desc == "Some category description"
    assert category.category_type == "expense"
    assert category.tax_schedule_info == "Tax schedule info"
    assert category.budget_amount == Decimal("123.45")


def test_from_list_with_custom_fields():
    """Test creating a category from a list of QIF strings"""
    setattr(Category, "__CUSTOM_FIELDS", [])  # Reset custom fields
    qif_list = [
        "NParent",
        "DSome category description",
        "E",
        "T",
        "B123.45",
        "RTax schedule info",
        "XCustom field 1",
        "Y9238479",
        "DT2022-01-01T00:00:00.000001",
    ]

    # Add custom fields
    Category.add_custom_field(
        line_code="X",
        attr="custom_field_1",
        field_type=str,
    )
    Category.add_custom_field(
        line_code="Y",
        attr="custom_field_2",
        field_type=Decimal,
    )
    Category.add_custom_field(
        line_code="DT",  # Test multi-character line code
        attr="custom_field_3",
        field_type=datetime,
    )

    category = Category.from_list(qif_list)
    assert category.name == "Parent"
    assert category.desc == "Some category description"
    assert category.category_type == "expense"
    assert category.tax_schedule_info == "Tax schedule info"
    assert category.budget_amount == Decimal("123.45")
    assert category.custom_field_1 == "Custom field 1"  # type: ignore
    assert category.custom_field_2 == Decimal("9238479")  # type: ignore
    assert category.custom_field_3 == datetime(2022, 1, 1, 0, 0, 0, 1)  # type: ignore
    setattr(Category, "__CUSTOM_FIELDS", [])  # Reset custom fields


def test_from_list_with_unknown_line_code():
    """Test creating a category from a list of QIF strings with an unknown
    line code
    """
    qif_list = [
        "NParent",
        "DSome category description",
        "E",
        "T",
        "B123.45",
        "RTax schedule info",
        "ZInvalid field",
    ]

    with pytest.raises(ValueError):
        Category.from_list(qif_list)


def test_from_string_default_separator():
    """Test creating a category from a QIF string"""
    qif_string = (
        "NParent\nDSome category description\nE\nT\nB123.45\nRTax schedule info\n"
    )
    category = Category.from_string(qif_string)
    assert category.name == "Parent"
    assert category.desc == "Some category description"
    assert category.category_type == "expense"
    assert category.tax_schedule_info == "Tax schedule info"
    assert category.budget_amount == Decimal("123.45")


def test_from_string_custom_separator():
    """Test creating a category from a QIF string with a custom separator"""
    qif_string = (
        "NParent---DSome category description---E---T---B123.45---RTax schedule info---"
    )
    category = Category.from_string(qif_string, separator="---")
    assert category.name == "Parent"
    assert category.desc == "Some category description"
    assert category.category_type == "expense"
    assert category.tax_schedule_info == "Tax schedule info"
    assert category.budget_amount == Decimal("123.45")


def test_create_categories_from_hierarchy():
    """Test creating categories from a hierarchy string"""
    hierarchy = "Parent:Child:Grandchild"
    grandchild = create_categories_from_hierarchy(hierarchy)
    assert grandchild.name == "Grandchild"
    assert grandchild.parent is not None
    assert grandchild.parent.name == "Child"
    assert grandchild.parent.parent is not None
    assert grandchild.parent.parent.name == "Parent"


def test_create_categories_from_hierarchy_single_category():
    """Test creating a single category from a hierarchy string"""
    hierarchy = "Parent"
    parent = create_categories_from_hierarchy(hierarchy)
    assert parent == Category(name="Parent")


def test_add_categories_to_container_no_existing_categories_list():
    """Test that a category is added to an empty list of categories."""
    new_category = Category(name="New Category")
    categories = []
    add_categories_to_container(new_category, categories)
    assert categories == [new_category]


def test_add_categories_to_container_with_hierarchy_parent_not_in_list():
    """Test that a category is added to a list of categories with hierarchy."""
    child = Category(name="Child")
    child.set_parent(Category(name="Parent"))
    categories = []
    add_categories_to_container(child, categories)
    assert categories == [
        Category(name="Parent", children=[Category(name="Child")]),
    ]


def test_add_categories_to_container_with_hierarchy_parent_in_list():
    """Test that a category is added to a list of categories with hierarchy."""
    parent = Category(name="Parent")
    child = Category(name="Child")
    child.set_parent(parent)
    categories = [parent]
    add_categories_to_container(child, categories)
    assert categories == [Category(name="Parent", children=[child])]


def test_add_categories_to_container_with_existing_categories_list():
    """Test that a category is added to a filled list of categories."""
    new_category = Category(name="New Category")
    existing_category = Category(name="Existing Category")
    categories = [existing_category]
    add_categories_to_container(new_category, categories)
    assert categories == [existing_category, new_category]


def test_add_categories_to_container_with_hierarchy_existing_categories_list():
    """Test that a category is added to a filled list of categories."""
    child = Category(name="Child")
    child.set_parent(Category(name="Parent"))
    existing_category = Category(name="Existing Category")
    categories = [existing_category]
    add_categories_to_container(child, categories)
    assert categories == [
        existing_category,
        Category(name="Parent", children=[Category(name="Child")]),
    ]


def test_add_categories_to_container_complex_hierarchy_list():
    """Test that a complex category hierarchy is added to a list of
    categories."""
    plant = Category(name="Plant")
    tree = Category(name="Tree")
    tree.set_parent(plant)

    oak = Category(name="Oak")
    oak.set_parent(tree)
    oak_leaf = Category(name="Oak Leaf")
    oak_leaf.set_parent(oak)

    # Create separate list of categories with a root with the same name
    tree2 = Category(name="Tree")
    birch = Category(name="Birch")
    birch.set_parent(tree2)
    birch_leaf = Category(name="Birch Leaf")
    birch_leaf.set_parent(birch)

    categories = [plant]
    add_categories_to_container(birch_leaf, categories)

    assert len(categories) == 1
    plant_traversal = categories[0].traverse_down()
    assert plant_traversal == [plant, tree, birch, birch_leaf, oak, oak_leaf]

    plant_render = categories[0].render_tree()
    assert plant_render == (
        "Plant (root)\n"
        "└─ Tree\n"
        "   └─ Oak\n"
        "      └─ Oak Leaf\n"
        "   └─ Birch\n"
        "      └─ Birch Leaf"
    )


def test_add_categories_to_container_no_existing_categories_dict():
    """Test that a category is added to an empty dict of categories."""
    new_category = Category(name="New Category")
    categories = {}
    add_categories_to_container(new_category, categories)
    assert categories == {new_category.name: new_category}


def test_add_categories_to_container_with_hierarchy_parent_not_in_dict():
    """Test that a category is added to a dict of categories with hierarchy."""
    child = Category(name="Child")
    child.set_parent(Category(name="Parent"))
    categories = {}
    add_categories_to_container(child, categories)
    assert categories == {
        "Parent": Category(name="Parent", children=[Category(name="Child")]),
    }


def test_add_categories_to_container_with_hierarchy_parent_in_dict():
    """Test that a category is added to a dict of categories with hierarchy."""
    parent = Category(name="Parent")
    child = Category(name="Child")
    child.set_parent(parent)
    categories = {parent.name: parent}
    add_categories_to_container(child, categories)
    assert categories == {"Parent": Category(name="Parent", children=[child])}


def test_add_categories_to_container_with_existing_categories_dict():
    """Test that a category is added to a filled dict of categories."""
    new_category = Category(name="New Category")
    existing_category = Category(name="Existing Category")
    categories = {existing_category.name: existing_category}
    add_categories_to_container(new_category, categories)
    assert categories == {
        existing_category.name: existing_category,
        new_category.name: new_category,
    }


def test_add_categories_to_container_with_hierarchy_existing_categories_dict():
    """Test that a category is added to a filled dict of categories."""
    child = Category(name="Child")
    child.set_parent(Category(name="Parent"))
    existing_category = Category(name="Existing Category")
    categories = {existing_category.name: existing_category}
    add_categories_to_container(child, categories)
    assert categories == {
        existing_category.name: existing_category,
        "Parent": Category(name="Parent", children=[Category(name="Child")]),
    }


def test_add_categories_to_container_complex_hierarchy_dict():
    """Test that a complex category hierarchy is added to a dict of
    categories."""
    plant = Category(name="Plant")
    tree = Category(name="Tree")
    tree.set_parent(plant)

    oak = Category(name="Oak")
    oak.set_parent(tree)
    oak_leaf = Category(name="Oak Leaf")
    oak_leaf.set_parent(oak)

    # Create separate list of categories with a root with the same name
    tree2 = Category(name="Tree")
    birch = Category(name="Birch")
    birch.set_parent(tree2)
    birch_leaf = Category(name="Birch Leaf")
    birch_leaf.set_parent(birch)

    categories = {plant.name: plant}
    add_categories_to_container(birch_leaf, categories)

    assert len(categories) == 1
    plant_traversal = categories[plant.name].traverse_down()
    assert plant_traversal == [plant, tree, birch, birch_leaf, oak, oak_leaf]

    plant_render = categories[plant.name].render_tree()
    assert plant_render == (
        "Plant (root)\n"
        "└─ Tree\n"
        "   └─ Oak\n"
        "      └─ Oak Leaf\n"
        "   └─ Birch\n"
        "      └─ Birch Leaf"
    )


def test_categories_do_not_override_each_other():
    """Test that categories do not override each other

    Related to issue #23.
    https://github.com/isaacharrisholt/quiffen/issues/23
    """
    test_file = Path(__file__).parent / "test_files" / "test_category_override.qif"

    qif = Qif.parse(test_file)

    assert len(qif.categories) == 2

    assert sorted(qif.categories.keys()) == ["Everyday Expenses", "Income"]

    income = qif.categories["Income"]
    assert income.name == "Income"
    assert len(income.children) == 2
    assert sorted(c.name for c in income.children) == [
        "Available next month",
        "Available this month",
    ]

    everyday_expenses = qif.categories["Everyday Expenses"]
    assert everyday_expenses.name == "Everyday Expenses"
    assert len(everyday_expenses.children) == 2
    assert sorted(c.name for c in everyday_expenses.children) == [
        "Food Budget",
        "Gousto",
    ]


def test_child_categories_inherit_category_type():
    """Test that child categories inherit the category type of their parent.

    Related to issue #37.
    https://github.com/isaacharrisholt/quiffen/issues/37
    """
    test_file = Path(__file__).parent / "test_files" / "test_category_inherit.qif"

    qif = Qif.parse(test_file)

    assert len(qif.categories) == 1

    root = qif.categories["Root"]
    assert root.name == "Root"
    assert root.category_type == CategoryType.INCOME
    assert len(root.children) == 1

    child = root.children[0]
    assert child.name == "Child"
    assert child.category_type == CategoryType.INCOME



from datetime import datetime
from decimal import Decimal

import pytest

from quiffen import Category
from quiffen.core.class_type import Class


def test_create_class():
    """Test creating a class"""
    cls = Class(name="Test")
    assert cls.name == "Test"
    assert cls.desc is None

    cls2 = Class(name="Test2", desc="Test Description")
    assert cls2.name == "Test2"
    assert cls2.desc == "Test Description"


def test_eq_success():
    """Test that two classes are equal"""
    cls = Class(name="Test")
    cls2 = Class(name="Test")
    assert cls == cls2


def test_eq_failure():
    """Test that two classes are not equal"""
    cls = Class(name="Test")
    cls2 = Class(name="Test2")
    assert cls != cls2


def test_str_method():
    """Test the string representation of a class"""
    cls = Class(name="Test")
    assert str(cls) == "Class:\n\tName: Test\n\tCategories: 0"

    cls2 = Class(name="Test2", desc="Test Description")
    assert str(cls2) == (
        "Class:\n\tName: Test2\n\tDescription: Test Description\n\tCategories: 0"
    )


def test_to_dict():
    """Test the to_dict method"""
    cls = Class(name="Test")
    assert cls.to_dict() == {"name": "Test", "desc": None, "categories": []}

    cls2 = Class(name="Test2", desc="Test Description")
    assert cls2.to_dict() == {
        "name": "Test2",
        "desc": "Test Description",
        "categories": [],
    }


def test_to_dict_with_ignore():
    """Test the to_dict method with ignore"""
    cls = Class(name="Test", desc="Test Description")
    assert cls.to_dict(ignore={"desc", "categories"}) == {"name": "Test"}


def test_add_category():
    """Test adding a category to a class"""
    cls = Class(name="Test")
    assert not cls.categories

    category = Category(name="Test Category")
    cls.add_category(category)
    assert cls.categories == [category]


def test_add_existing_category():
    """Test adding an existing category to a class"""
    cls = Class(name="Test")
    category = Category(name="Test Category")
    cls.add_category(category)
    assert cls.categories == [category]

    cls.add_category(category)
    assert cls.categories == [category]


def test_merge():
    """Test merging two classes"""
    category1 = Category(name="Category 1")
    category2 = Category(name="Category 2")
    cls = Class(name="Test", categories=[category1])
    cls2 = Class(name="Test2", desc="Test Description", categories=[category2])
    cls.merge(cls2)
    assert cls.name == "Test"
    assert cls.desc == "Test Description"
    assert cls.categories == [category1, category2]


def test_to_qif():
    """Test the to_qif method"""
    cls = Class(
        name="Test Class",
        desc="Test Description",
    )
    assert cls.to_qif() == "!Type:Class\nNTest Class\nDTest Description\n"


def test_from_list_no_custom_fields():
    """Test creating a class from a list of QIF strings"""
    qif_list = [
        "NTest",
        "DTest Description",
    ]
    cls = Class.from_list(qif_list)
    assert cls.name == "Test"
    assert cls.desc == "Test Description"


def test_from_list_with_custom_fields():
    """Test creating a class from a list of QIF strings"""
    setattr(Class, "__CUSTOM_FIELDS", [])  # Reset custom fields
    qif_list = [
        "NTest",
        "DTest Description",
        "XCustom field 1",
        "Y9238479",
        "DT2022-01-01T00:00:00.000001",
    ]

    # Add custom fields
    Class.add_custom_field(
        line_code="X",
        attr="custom_field_1",
        field_type=str,
    )
    Class.add_custom_field(
        line_code="Y",
        attr="custom_field_2",
        field_type=Decimal,
    )
    Class.add_custom_field(
        line_code="DT",  # Test multi-character line code
        attr="custom_field_3",
        field_type=datetime,
    )

    cls = Class.from_list(qif_list)
    assert cls.name == "Test"
    assert cls.desc == "Test Description"
    assert cls.custom_field_1 == "Custom field 1"  # type: ignore
    assert cls.custom_field_2 == Decimal("9238479")  # type: ignore
    assert cls.custom_field_3 == datetime(2022, 1, 1, 0, 0, 0, 1)  # type: ignore
    setattr(Class, "__CUSTOM_FIELDS", [])  # Reset custom fields


def test_from_list_with_unknown_line_code():
    """Test creating a class from a list of QIF strings with an unknown line
    code"""
    qif_list = [
        "NTest",
        "DTest Description",
        "ZCustom field 1",
    ]

    with pytest.raises(ValueError):
        Class.from_list(qif_list)


def test_from_string_default_separator():
    """Test creating a class from a QIF string"""
    qif_string = "NTest\nDTest Description\n"
    cls = Class.from_string(qif_string)
    assert cls.name == "Test"
    assert cls.desc == "Test Description"


def test_from_string_custom_separator():
    """Test creating a class from a QIF string with a custom separator"""
    qif_string = "NTest---DTest Description---"
    cls = Class.from_string(qif_string, separator="---")
    assert cls.name == "Test"
    assert cls.desc == "Test Description"



from datetime import datetime
from decimal import Decimal

import pytest

from quiffen import Account, Transaction
from quiffen.core.account import AccountType


def test_create_account():
    """Test creating an account"""
    account = Account(name="Test Account")
    assert account.name == "Test Account"
    assert account.account_type is None
    assert account.desc is None
    assert account.balance is None
    assert not account.transactions
    assert account.date_at_balance is None
    assert account.credit_limit is None

    account2 = Account(
        name="Test Account",
        account_type=AccountType.BANK,
        desc="Test Description",
        balance=Decimal(100),
    )
    assert account2.name == "Test Account"
    assert account2.account_type == "Bank"
    assert account2.desc == "Test Description"
    assert account2.balance == 100
    assert not account2.transactions

    account3 = Account(
        name="Test Account",
        account_type=AccountType.BANK,
        desc="Test Description",
        balance=Decimal(100),
        transactions={
            AccountType.BANK: [
                Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
            ],
        },
    )
    assert account3.name == "Test Account"
    assert account3.account_type == "Bank"
    assert account3.desc == "Test Description"
    assert account3.balance == 100
    assert account3.transactions == {
        "Bank": [Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))],
    }


def test_eq_success():
    """Test comparing two accounts for equality"""
    account1 = Account(name="Test Account")
    account2 = Account(name="Test Account")
    assert account1 == account2


def test_eq_failure():
    """Test comparing two accounts for equality"""
    account1 = Account(name="Test Account")
    account2 = Account(name="Test Account 2")
    assert account1 != account2

    account3 = Account(name="Test Account", account_type=AccountType.CREDIT_CARD)
    account4 = Account(name="Test Account", account_type=AccountType.BANK)
    assert account3 != account4


def test_str_method():
    """Test the string representation of an account"""
    account = Account(
        name="Test Account",
        account_type=AccountType.BANK,
        desc="Test Description",
        balance=Decimal(100),
        transactions={
            AccountType.BANK: [
                Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
            ],
        },
    )
    assert str(account) == (
        "Account:\n\tName: Test Account\n\tDesc: Test Description\n\t"
        "Account Type: Bank\n\tBalance: 100\n\tTransactions: 1"
    )

    account.add_transaction(
        Transaction(date=datetime(2022, 2, 1), amount=Decimal(0)),
        header=AccountType.BANK,
    )
    assert str(account) == (
        "Account:\n\tName: Test Account\n\tDesc: Test Description\n\t"
        "Account Type: Bank\n\tBalance: 100\n\tTransactions: 2"
    )

    account.add_transaction(
        Transaction(date=datetime(2022, 2, 1), amount=Decimal(0)),
        header=AccountType.CREDIT_CARD,
    )
    assert str(account) == (
        "Account:\n\tName: Test Account\n\tDesc: Test Description\n\t"
        "Account Type: Bank\n\tBalance: 100\n\tTransactions: 3"
    )


# pylint: disable=protected-access
def test_set_header():
    """Test setting the header for an account"""
    account = Account(name="Test Account")
    assert account._last_header is None
    account.set_header(AccountType.BANK)
    assert account._last_header == "Bank"


# pylint: enable=protected-access


def test_add_transaction():
    """Test adding a transaction to an account"""
    account = Account(name="Test Account")
    transaction = Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
    account.add_transaction(transaction, header=AccountType.BANK)
    assert account.transactions == {
        "Bank": [Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))],
    }


def test_add_transaction_invalid_header():
    """Test adding a transaction to an account with an invalid header"""
    account = Account(name="Test Account")
    transaction = Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
    with pytest.raises(ValueError):
        account.add_transaction(transaction, header="Invalid")  # type: ignore


def test_add_transaction_no_header():
    """Test adding a transaction to an account with no header"""
    account = Account(name="Test Account")
    transaction = Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
    with pytest.raises(RuntimeError):
        account.add_transaction(transaction)


def test_merge():
    """Test merging two accounts"""
    account1 = Account(
        name="Test Account",
        account_type=AccountType.BANK,
        balance=Decimal(100),
        transactions={
            AccountType.BANK: [
                Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
            ],
        },
    )
    account2 = Account(
        name="Test Account",
        account_type=AccountType.BANK,
        desc="Test Description",
        balance=Decimal(100),
        transactions={
            AccountType.BANK: [
                Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
            ],
        },
    )
    account1.merge(account2)
    assert account1.name == "Test Account"
    assert account1.desc == "Test Description"
    assert account1.transactions == {
        AccountType.BANK: [
            Transaction(date=datetime(2022, 2, 1), amount=Decimal(0)),
            Transaction(date=datetime(2022, 2, 1), amount=Decimal(0)),
        ],
    }


def test_to_qif():
    """Test converting an account to QIF"""
    account = Account(
        name="Test Account",
        account_type=AccountType.BANK,
        desc="Test Description",
        balance=Decimal(100),
        transactions={
            AccountType.BANK: [
                Transaction(date=datetime(2022, 2, 1), amount=Decimal(0))
            ],
        },
    )
    account_qif = account.to_qif()
    print(repr(account_qif))
    assert account_qif == (
        "!Account\nNTest Account\nDTest Description\nTBank\n$100\n"
        "^\n!Type:Bank\nD2022-02-01\nT0\n"
    )


def test_from_list_no_custom_fields():
    """Test creating an account from a list of sections with no custom fields"""
    qif_list = [
        "!This should be ignored",
        "NTest Account",
        "DTest Description",
        "TCCard",
        "L1000",
        "$100",
        "/2022-02-01",
    ]
    account = Account.from_list(qif_list)
    assert account.name == "Test Account"
    assert account.desc == "Test Description"
    assert account.account_type == "CCard"
    assert account.credit_limit == 1000
    assert account.balance == 100
    assert account.date_at_balance == datetime(2022, 2, 1)
    assert not account.transactions


def test_from_list_with_custom_fields():
    """Test creating an account from a list of sections with custom fields"""
    setattr(Account, "__CUSTOM_FIELDS", [])  # Reset custom fields
    qif_list = [
        "!This should be ignored",
        "NTest Account",
        "DTest Description",
        "TCCard",
        "L1000",
        "$100",
        "/2022-02-01",
        "XCustom field 1",
        "Y9238479",
        "DT2022-01-01T00:00:00.000001",
    ]

    # Add custom fields
    Account.add_custom_field(
        line_code="X",
        attr="custom_field_1",
        field_type=str,
    )
    Account.add_custom_field(
        line_code="Y",
        attr="custom_field_2",
        field_type=Decimal,
    )
    Account.add_custom_field(
        line_code="DT",  # Test multi-character line code
        attr="custom_field_3",
        field_type=datetime,
    )

    account = Account.from_list(qif_list)
    assert account.name == "Test Account"
    assert account.desc == "Test Description"
    assert account.account_type == "CCard"
    assert account.credit_limit == 1000
    assert account.balance == 100
    assert account.date_at_balance == datetime(2022, 2, 1)
    assert not account.transactions
    assert account.custom_field_1 == "Custom field 1"  # type: ignore
    assert account.custom_field_2 == Decimal("9238479")  # type: ignore
    assert account.custom_field_3 == datetime(2022, 1, 1, 0, 0, 0, 1)  # type: ignore
    setattr(Account, "__CUSTOM_FIELDS", [])  # Reset custom fields


def test_from_list_with_unknown_line_code():
    """Test creating an account from a list of QIF strings with an unknown
    line code
    """
    qif_list = [
        "!This should be ignored",
        "NTest Account",
        "DTest Description",
        "TCCard",
        "L1000",
        "$100",
        "/2022-02-01",
        "ZInvalid field",
    ]

    with pytest.raises(ValueError):
        Account.from_list(qif_list)


def test_from_string_default_separator():
    """Test creating an account from a string with the default separator"""
    qif_string = (
        "!This should be ignored\n"
        "NTest Account\n"
        "DTest Description\n"
        "TCCard\n"
        "L1000\n"
        "$100\n"
        "/2022-02-01\n"
    )
    account = Account.from_string(qif_string)
    assert account.name == "Test Account"
    assert account.desc == "Test Description"
    assert account.account_type == "CCard"
    assert account.credit_limit == 1000
    assert account.balance == 100
    assert account.date_at_balance == datetime(2022, 2, 1)
    assert not account.transactions


def test_from_string_custom_separator():
    """Test creating an account from a string with a custom separator"""
    qif_string = (
        "!This should be ignored---"
        "NTest Account---"
        "DTest Description---"
        "TCCard---"
        "L1000---"
        "$100---"
        "/2022-02-01---"
    )
    account = Account.from_string(qif_string, separator="---")
    assert account.name == "Test Account"
    assert account.desc == "Test Description"
    assert account.account_type == "CCard"
    assert account.credit_limit == 1000
    assert account.balance == 100
    assert account.date_at_balance == datetime(2022, 2, 1)
    assert not account.transactions


def test_to_dict():
    """Test converting an account to a dictionary"""
    account = Account(name="Test Account")
    account_dict = account.to_dict()
    assert account_dict == {
        "name": "Test Account",
        "desc": None,
        "account_type": None,
        "credit_limit": None,
        "balance": None,
        "date_at_balance": None,
        "transactions": {},
    }


def test_to_dict_with_ignore():
    """Test converting an account to a dictionary with ignored fields"""
    account = Account(name="Test Account")
    account_dict = account.to_dict(ignore=["name", "desc"])
    assert account_dict == {
        "account_type": None,
        "credit_limit": None,
        "balance": None,
        "date_at_balance": None,
        "transactions": {},
    }



from datetime import datetime

import pytest

from quiffen import utils


@pytest.mark.parametrize(
    "date_pattern,day_first",
    [
        # Day first patterns
        ("%d/%m/%Y", True),
        ("%d-%m-%Y", True),
        ("%d/%m/%y", True),
        ("%d-%m-%y", True),
        ("%d0%B0%Y", True),
        ("%d0%B0%y", True),
        ("%d0%b0%Y", True),
        ("%d0%b0%y", True),
        # Month first patterns
        ("%m/%d/%Y", False),
        ("%m-%d-%Y", False),
        ("%m/%d/%y", False),
        ("%m-%d-%y", False),
        ("%B0%d0%Y", False),
        ("%B0%d0%y", False),
        ("%b0%d0%Y", False),
        ("%b0%d0%y", False),
        # Year first patterns
        ("%Y/%m/%d", False),
        ("%Y-%m-%d", False),
        ("%Y0%B0%d", False),
        ("%Y0%b0%d", False),
    ],
)
def test_parse_date(date_pattern, day_first):
    want = datetime(2022, 1, 2)
    input_string = want.strftime(date_pattern)
    got = utils.parse_date(input_string, day_first)
    assert got == want


@pytest.mark.parametrize(
    "field,expected_line_code,expected_field_info",
    [
        ("DDate\n\n", "D", "Date"),
        ("PPayee", "P", "Payee"),
        ("MMemo", "M", "Memo"),
        ("TAmount", "T", "Amount"),
        ("LCategory", "L", "Category"),
        ("SSplit", "S", "Split"),
        ("E", "E", ""),
        ("", "", ""),
    ],
)
def test_parse_line_code_and_field_info(
    field,
    expected_line_code,
    expected_field_info,
):
    line_code, field_info = utils.parse_line_code_and_field_info(field)
    assert line_code == expected_line_code
    assert field_info == expected_field_info


def test_apply_csv_formatting_scalar():
    assert utils.apply_csv_formatting_to_scalar("foo") == "foo"
    assert utils.apply_csv_formatting_to_scalar(123) == 123
    assert utils.apply_csv_formatting_to_scalar(123.45) == 123.45
    assert utils.apply_csv_formatting_to_scalar(123.0) == 123
    assert (
        utils.apply_csv_formatting_to_scalar(
            datetime(2022, 1, 2),
        )
        == "2022-01-02"
    )
    assert (
        utils.apply_csv_formatting_to_scalar(
            datetime(2022, 1, 2).date(),
        )
        == "2022-01-02"
    )


def test_apply_csv_formatting_to_container_list():
    assert utils.apply_csv_formatting_to_container(
        ["foo", 123, 123.45, 123.0, datetime(2022, 1, 2)],
    ) == ["foo", 123, 123.45, 123, "2022-01-02"]


def test_apply_csv_formatting_to_container_dict():
    assert utils.apply_csv_formatting_to_container(
        {
            "foo": 123,
            "bar": 123.45,
            "baz": 123.0,
            "qux": datetime(2022, 1, 2),
        },
    ) == {
        "foo": 123,
        "bar": 123.45,
        "baz": 123,
        "qux": "2022-01-02",
    }


def test_apply_csv_formatting_to_container_dict_with_dicts_and_lists():
    assert utils.apply_csv_formatting_to_container(
        {
            "foo": 123,
            "bar": 123.45,
            "baz": 123.0,
            "qux": datetime(2022, 1, 2),
            "quux": {
                "foo": 123,
                "bar": 123.45,
                "baz": 123.0,
                "qux": datetime(2022, 1, 2),
            },
            "quuz": [
                123,
                123.45,
                123.0,
                datetime(2022, 1, 2),
            ],
        },
    ) == {
        "foo": 123,
        "bar": 123.45,
        "baz": 123,
        "qux": "2022-01-02",
        "quux": {
            "foo": 123,
            "bar": 123.45,
            "baz": 123,
            "qux": "2022-01-02",
        },
        "quuz": [
            123,
            123.45,
            123,
            "2022-01-02",
        ],
    }



from __future__ import annotations

import csv
import io
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from pydantic.types import FilePath

from quiffen import utils
from quiffen.core.account import Account, AccountType
from quiffen.core.base import BaseModel, Field
from quiffen.core.category import Category, add_categories_to_container
from quiffen.core.class_type import Class
from quiffen.core.investment import Investment
from quiffen.core.security import Security
from quiffen.core.transaction import Transaction

if TYPE_CHECKING:
    import pandas as pd

VALID_TRANSACTION_ACCOUNT_TYPES = [
    "!type:cash",
    "!type:bank",
    "!type:ccard",
    "!type:otha",
    "!type:othl",
    "!type:invoice",
]


class QifDataType(str, Enum):
    """An Enum representing the different types of data that can be found in a
    Qif object."""

    TRANSACTIONS = "transactions"
    INVESTMENTS = "investments"
    CLASSES = "classes"
    CATEGORIES = "categories"
    ACCOUNTS = "accounts"
    SECURITIES = "securities"


class ParserException(Exception):
    pass


class Qif(BaseModel):
    """
    The main class of the package. For parsing QIF files.

    See the readme for usage examples.

    Parameters
    ----------
    accounts : dict, default=None
        A dict of accounts in the form {'Account Name': account_object}.
    categories : dict, default=None
        A dict of categories in the form {'Category Name': category_object}.
    classes : dict, default=None
        A dict of classes in the form {'Class Name': class_object}.
    """

    accounts: dict[str, Account] = {}
    categories: dict[str, Category] = {}
    classes: dict[str, Class] = {}
    securities: dict[str, Security] = {}

    __CUSTOM_FIELDS: list[Field] = []  # type: ignore

    def __str__(self) -> str:
        accounts_str = "\n".join(str(acc) for acc in self.accounts.values())
        categories_str = "\n".join(str(cat) for cat in self.categories.values())
        classes_str = "\n".join(str(cls) for cls in self.classes.values())
        securities_str = "\n".join(str(sec) for sec in self.securities.values())

        if not (accounts_str or categories_str or classes_str):
            return "Empty Qif object"

        return_str = "QIF\n===\n\n"

        if accounts_str:
            return_str += f"Accounts\n--------\n\n{accounts_str}\n\n"
        if categories_str:
            return_str += f"Categories\n----------\n\n{categories_str}\n\n"
        if classes_str:
            return_str += f"Classes\n-------\n\n{classes_str}\n\n"
        if securities_str:
            return_str += f"Securities\n----------\n\n{securities_str}\n\n"

        return return_str

    @classmethod
    def from_list(cls, lst: list[str]) -> Qif:
        raise NotImplementedError(
            "This method is not implemented for Qif objects. Use Qif.parse to "
            "parse a QIF file."
        )

    @classmethod
    def parse(
        cls,
        path: Union[FilePath, str],
        separator: str = "\n",
        day_first: bool = False,
        encoding: str = "utf-8",
    ) -> Qif:
        """Return a class instance from a QIF file.

        Parameters
        ----------
        path : Union[FilePath, str]
            The path to the QIF file.
        separator : str, default='\n'
            The line separator for the QIF file. This probably won't need
            changing.
        day_first : bool, default=False
            Whether the day or month comes first in the date.
        encoding : str, default='utf-8'
            The encoding of the QIF file.

        Returns
        -------
        Qif
            A Qif object containing all the data in the QIF file.
        """
        path = Path(path)
        if path.suffix.lower() != ".qif":
            raise ParserException("The file must be a QIF file.")

        if not path.exists():
            raise ParserException("The file does not exist.")

        data = path.read_text(encoding=encoding).strip().strip("\n")

        if not data:
            raise ParserException("The file is empty.")

        return cls.parse_string(data, separator=separator, day_first=day_first)

    @classmethod
    def parse_string(
        cls, data: str, separator: str = "\n", day_first: bool = False
    ) -> Qif:
        """Return a class instance from a string containing QIF data.

        Parameters
        ----------
        data : str
            The string of QIF data
        separator : str, default='\n'
            The line separator for the QIF file. This probably won't need
            changing.
        day_first : bool, default=False
            Whether the day or month comes first in the date.

        Returns
        -------
        Qif
            A Qif object containing all the data in the QIF file.
        """
        if not data:
            raise ParserException("The data string is empty.")

        accounts: dict[str, Account] = {}
        last_account = None
        categories: dict[str, Category] = {}
        classes: dict[str, Class] = {}
        securities: dict[str, Security] = {}

        sections = data.split("^")
        last_header = None
        line_number = 1

        for section in sections:
            if not section:
                continue

            section_lines = section.split(separator)
            if not section_lines:
                continue

            # Allow for comments and blank lines at the top of sections
            # Also skip `!Clear:` lines to ignore flags such as `!Clear:AutoSwitch`
            # Also skip `!Option:` lines to ignore flags such as `!Option:AutoSwitch`
            for i, line in enumerate(section_lines):
                if (
                    not line.strip()
                    or line.strip().startswith("!Clear:")
                    or line.strip().startswith("!Option:")
                    or line[0] == "#"
                ):
                    continue

                # Found the header line
                header_line = line
                line_number += i
                section_lines = section_lines[i:]
                break
            else:
                # Empty section
                continue

            if header_line[0] != "!":
                if not last_header:
                    raise ParserException(
                        f"Line {line_number}: " f"No header found before transactions."
                    )
                header_line = last_header

            last_header = header_line

            sanitised_section_lines = [
                line
                for line in section_lines
                if line.strip() and line.strip()[0] != "!"
            ]

            if not sanitised_section_lines:
                continue

            # Check for new categories and accounts first, otherwise it's a
            # transaction so a default account is created
            if "!Type:Cat" in header_line:
                # Section contains category information
                new_category = Category.from_list(sanitised_section_lines)
                categories = add_categories_to_container(
                    new_category,
                    categories,
                )
            elif "!Type:Class" in header_line:
                new_class = Class.from_list(sanitised_section_lines)
                if new_class.name in classes:
                    classes[new_class.name].merge(new_class)
                else:
                    classes[new_class.name] = new_class
            elif "!Account" in header_line:
                new_account = Account.from_list(sanitised_section_lines)
                if new_account.name in accounts:
                    accounts[new_account.name].merge(new_account)
                else:
                    accounts[new_account.name] = new_account
                last_account = new_account.name
            elif "!Type:Invst" in header_line:
                # Investment
                new_investment = Investment.from_list(
                    sanitised_section_lines,
                    day_first=day_first,
                    line_number=line_number,
                )

                if last_account is None:
                    raise ParserException(
                        f"Line {line_number}: "
                        "No account found before investment. "
                        "This should not happen."
                    )

                accounts[last_account].add_transaction(
                    new_investment,
                    AccountType("Invst"),
                )
            elif "!Type:Security" in header_line:
                # Security
                new_security = Security.from_list(
                    sanitised_section_lines,
                    line_number=line_number,
                )
                if new_security.symbol is None:
                    raise ParserException(
                        f"Line {line_number}: " f"No symbol found for security."
                    )
                securities[new_security.symbol] = new_security
            elif "!Type" in header_line and not accounts:
                # Accounts is empty and there's a transaction, so create default
                # account to put transactions in
                default_account = Account(
                    name="Quiffen Default Account",
                    desc=(
                        "The default account created by Quiffen when no other "
                        "accounts were present"
                    ),
                )
                accounts[default_account.name] = default_account
                last_account = default_account.name

            if header_line.lower().replace(" ", "") in (
                VALID_TRANSACTION_ACCOUNT_TYPES
            ):
                # Other transaction type
                new_transaction, new_classes = Transaction.from_list(
                    sanitised_section_lines,
                    day_first=day_first,
                    line_number=line_number,
                )

                if last_account is None:
                    raise ParserException(
                        f"Line {line_number}: "
                        "No account found before transactions. "
                        "This should not happen."
                    )

                accounts[last_account].add_transaction(
                    new_transaction,
                    AccountType(header_line.split(":")[1].strip()),
                )

                if new_transaction.category:
                    root = new_transaction.category.traverse_up()[-1]
                    if root.name in categories:
                        categories[root.name].merge(root)
                    else:
                        categories[root.name] = root

                for class_name, new_class in new_classes.items():
                    if class_name in classes:
                        classes[class_name].merge(new_class)
                    else:
                        classes[class_name] = new_class

            line_number += len(section.split("\n"))

        return cls(
            accounts=accounts,
            categories=categories,
            classes=classes,
            securities=securities,
        )

    def add_account(self, new_account: Account) -> None:
        """Add a new account to the Qif object"""
        if new_account.name in self.accounts:
            self.accounts[new_account.name].merge(new_account)
        else:
            self.accounts[new_account.name] = new_account

    def remove_account(self, account_name: str) -> Account:
        """Remove an account from this Qif object"""
        try:
            return self.accounts.pop(account_name)
        except KeyError as e:
            raise KeyError(
                f'Account "{account_name}" does not exist in this Qif object.'
            ) from e

    def add_category(self, new_category: Category) -> None:
        """Add a new category to the Qif object"""
        self.categories = add_categories_to_container(
            new_category,
            self.categories,
        )

    def remove_category(
        self,
        category_name: str,
        keep_children: bool = True,
    ) -> Category:
        """Remove a category from this Qif object"""
        try:
            category = self.categories.pop(category_name)
        except KeyError as e:
            raise KeyError(
                f'Category "{category_name}" does not exist in this Qif object.'
            ) from e

        if keep_children:
            print("Keeping children")
            for child in category.children:
                child.set_parent(None)
                self.add_category(child)

        return category

    def add_class(self, new_class: Class) -> None:
        """Add a new class to the Qif object"""
        if new_class.name in self.classes:
            self.classes[new_class.name].merge(new_class)
        else:
            self.classes[new_class.name] = new_class

    def remove_class(self, class_name: str) -> Class:
        """Remove a class from this Qif object"""
        try:
            return self.classes.pop(class_name)
        except KeyError as e:
            raise KeyError(
                f'Class "{class_name}" does not exist in this Qif object.'
            ) from e

    def add_security(self, new_security: Security) -> None:
        """Add a new security to the Qif object"""
        if not new_security.symbol:
            raise ValueError(
                "Cannot add a security without a symbol to the Qif object."
            )

        if new_security.symbol in self.securities:
            self.securities[new_security.symbol].merge(new_security)
        else:
            self.securities[new_security.symbol] = new_security

    def remove_security(self, security_symbol: str) -> Security:
        """Remove a security from this Qif object"""
        try:
            return self.securities.pop(security_symbol)
        except KeyError as e:
            raise KeyError(
                f'Security "{security_symbol}" does not exist in this Qif ' f"object."
            ) from e

    def to_qif(
        self,
        path: Optional[Union[FilePath, str, None]] = None,
        date_format: str = "%Y-%m-%d",
        encoding: str = "utf-8",
    ) -> str:
        """Convert the Qif object to a QIF file"""
        qif = ""

        if self.categories:
            qif += (
                "^\n".join(category.to_qif() for category in self.categories.values())
                + "^\n"
            )

        if self.classes:
            qif += "^\n".join(cls.to_qif() for cls in self.classes.values()) + "^\n"

        if self.accounts:
            qif += (
                "^\n".join(
                    account.to_qif(date_format=date_format, classes=self.classes)
                    for account in self.accounts.values()
                )
                + "^\n"
            )

        if path:
            Path(path).write_text(qif, encoding=encoding)

        return qif

    def _get_data_dicts(
        self,
        data_type: QifDataType = QifDataType.TRANSACTIONS,
        date_format: Optional[str] = "%Y-%m-%d",
        ignore: Optional[list[str]] = None,
    ) -> list[dict[str, Any]]:
        """Converts specified data from the Qif object to a list of dicts"""
        if ignore is None:
            ignore = []

        if data_type == QifDataType.TRANSACTIONS:
            data_dicts = []
            for account in self.accounts.values():
                for transaction_list in account.transactions.values():
                    data_dicts += [
                        transaction.to_dict(ignore=ignore)
                        for transaction in transaction_list
                        if not isinstance(transaction, Investment)
                    ]
        elif data_type == QifDataType.INVESTMENTS:
            data_dicts = []
            for account in self.accounts.values():
                for transaction_list in account.transactions.values():
                    data_dicts += [
                        transaction.to_dict(ignore=ignore)
                        for transaction in transaction_list
                        if isinstance(transaction, Investment)
                    ]
        elif data_type == QifDataType.ACCOUNTS:
            data_dicts = [
                account.to_dict(ignore=ignore) for account in self.accounts.values()
            ]
        elif data_type == QifDataType.CATEGORIES:
            data_dicts = [
                category.to_dict(ignore=ignore) for category in self.categories.values()
            ]
        elif data_type == QifDataType.CLASSES:
            data_dicts = [
                class_.to_dict(ignore=ignore) for class_ in self.classes.values()
            ]
        elif data_type == QifDataType.SECURITIES:
            data_dicts = [
                security.to_dict(ignore=ignore) for security in self.securities.values()
            ]
        else:
            raise ValueError(
                f"Invalid data_type: {data_type}. Must be one of "
                f"{list(QifDataType)}"
            )

        # Format and hide private fields
        return [  # type: ignore
            utils.apply_csv_formatting_to_container(
                {
                    k: v
                    for k, v in data_dict.items()
                    if not k.startswith("_") and k not in ignore
                },
                date_format=date_format,
            )
            for data_dict in data_dicts
        ]

    def to_csv(
        self,
        path: Optional[Union[FilePath, str, None]] = None,
        data_type: QifDataType = QifDataType.TRANSACTIONS,
        date_format: str = "%Y-%m-%d",
        ignore: Optional[list[str]] = None,
        delimiter: str = ",",
        quote_character: str = '"',
        encoding: str = "utf-8",
    ) -> str:
        """Convert part of the Qif object to a CSV file. The data_type
        parameter can be used to specify which part of the Qif object to
        convert to CSV. The default is to convert transactions.

        Parameters
        ----------
        path : Union[FilePath, str, None], optional
            The path to write the CSV file to, by default None
        data_type : QifDataType, optional
            The type of data to convert to CSV, by default
            QifDataType.TRANSACTIONS
        date_format : str, optional
            The date format to use when converting dates to strings, by
            default '%Y-%m-%d'
        ignore : list[str], optional
            A list of fields to ignore when converting to CSV, by default
            None
        delimiter : str, optional
            The delimiter to use when writing the CSV file, by default ','
        quote_character : str, optional
            The quote character to use when writing the CSV file, by
            default '"'
        encoding : str, default='utf-8'
            The encoding to use when writing the CSV file
        """
        if ignore is None:
            ignore = []

        data_dicts = self._get_data_dicts(
            data_type=data_type,
            date_format=date_format,
            ignore=ignore,
        )

        headers: set[str] = set()
        for data_dict in data_dicts:
            headers.update(k for k in data_dict.keys())

        output = io.StringIO()

        writer = csv.DictWriter(
            output,
            fieldnames=headers,
            extrasaction="ignore",  # Ignore extra fields (e.g. private fields)
            dialect="unix",  # Use Unix line endings
            delimiter=delimiter,
            quotechar=quote_character,
        )
        writer.writeheader()
        for data_dict in data_dicts:
            writer.writerow(data_dict)

        return_value = output.getvalue()

        if path:
            Path(path).write_text(return_value, encoding=encoding)

        return return_value

    def to_dataframe(
        self,
        data_type: QifDataType = QifDataType.TRANSACTIONS,
        ignore: Optional[list[str]] = None,
    ) -> pd.DataFrame:
        """Convert part of the Qif object to a Pandas DataFrame. The
        data_type parameter can be used to specify which part of the Qif
        object to convert to a DataFrame. The default is to convert
        transactions.

        Parameters
        ----------
        data_type : QifDataType, optional
            The type of data to convert to a DataFrame, by default
            QifDataType.TRANSACTIONS
        ignore : list[str], optional
            A list of fields to ignore when converting to a DataFrame, by
            default None

        Returns
        -------
        pd.DataFrame
            The data as a Pandas DataFrame
        """
        try:
            import pandas as pd
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "The pandas package is required to convert Qif objects to DataFrames. Please install it using `pip install quiffen[pandas]`."
            ) from e

        if ignore is None:
            ignore = []

        data_dicts = self._get_data_dicts(
            data_type=data_type,
            date_format=None,
            ignore=ignore,
        )

        return pd.DataFrame(data_dicts)



# pylint: disable=redefined-outer-name
import csv
from datetime import datetime
from decimal import Decimal
from pathlib import Path

import pytest

from quiffen import Investment, Transaction
from quiffen.core.account import Account, AccountType
from quiffen.core.category import Category
from quiffen.core.class_type import Class
from quiffen.core.qif import ParserException, Qif, QifDataType
from quiffen.core.security import Security


@pytest.fixture
def qif_file():
    return Path(__file__).parent / "test_files" / "test.qif"


@pytest.fixture
def qif_file_with_oth_a_account():
    return Path(__file__).parent / "test_files" / "test_oth_a.qif"


@pytest.fixture
def empty_file():
    return Path(__file__).parent / "test_files" / "empty.qif"


@pytest.fixture
def txt_file():
    return Path(__file__).parent / "test_files" / "invalid.txt"


@pytest.fixture
def nonexistent_file():
    return Path(__file__).parent / "test_files" / "nonexistent.qif"


@pytest.fixture
def qif_file_with_clear_autoswitch():
    return Path(__file__).parent / "test_files" / "test_clear_autoswitch.qif"


@pytest.fixture
def qif_file_with_option_autoswitch():
    return Path(__file__).parent / "test_files" / "test_option_autoswitch.qif"


@pytest.fixture
def qif_file_with_split_to_account():
    return Path(__file__).parent / "test_files" / "test_split.qif"


@pytest.fixture
def qif_file_with_unknown_account_type():
    return Path(__file__).parent / "test_files" / "test_unknown_account_type.qif"


@pytest.fixture
def qif_file_with_no_symbol():
    return Path(__file__).parent / "test_files" / "test_nosymbol.qif"


@pytest.fixture
def qif_file_circular_reference():
    return Path(__file__).parent / "test_files" / "test_circular_reference.qif"


def test_create_qif():
    """Test creating a Qif instance"""
    qif = Qif()
    assert not qif.accounts
    assert not qif.classes
    assert not qif.categories


def test_eq_success_empty_qif():
    """Test that two Qif instances are equal when empty"""
    qif = Qif()
    qif2 = Qif()
    assert qif == qif2


def test_eq_success_with_accounts():
    """Test that two Qif instances are equal when they have accounts"""
    qif = Qif()
    qif.add_account(Account(name="Test Account"))
    qif2 = Qif()
    qif2.add_account(Account(name="Test Account"))
    assert qif == qif2


def test_eq_success_with_classes():
    """Test that two Qif instances are equal when they have classes"""
    qif = Qif()
    qif.add_class(Class(name="Test Class"))
    qif2 = Qif()
    qif2.add_class(Class(name="Test Class"))
    assert qif == qif2


def test_eq_success_with_categories():
    """Test that two Qif instances are equal when they have categories"""
    qif = Qif()
    qif.add_category(Category(name="Test Category"))
    qif2 = Qif()
    qif2.add_category(Category(name="Test Category"))
    assert qif == qif2


def test_eq_failure_with_accounts():
    """Test that two Qif instances are not equal when they have different
    accounts"""
    qif = Qif()
    qif.add_account(Account(name="Test Account"))
    qif2 = Qif()
    qif2.add_account(Account(name="Test Account 2"))
    assert qif != qif2


def test_eq_failure_with_categories():
    """Test that two Qif instances are not equal when they have different
    categories"""
    qif = Qif()
    qif.add_category(Category(name="Test Category"))
    qif2 = Qif()
    qif2.add_category(Category(name="Test Category 2"))
    assert qif != qif2


def test_eq_failure_with_classes():
    """Test that two Qif instances are not equal when they have different
    classes"""
    qif = Qif()
    qif.add_class(Class(name="Test Class"))
    qif2 = Qif()
    qif2.add_class(Class(name="Test Class 2"))
    assert qif != qif2


def test_str_method_empty_qif():
    """Test the string representation of an empty Qif instance"""
    qif = Qif()
    assert str(qif) == "Empty Qif object"


def test_str_method_with_accounts():
    """Test the string representation of a Qif instance with accounts"""
    qif = Qif()
    qif.add_account(Account(name="Test Account"))
    assert str(qif) == (
        "QIF\n===\n\nAccounts\n--------\n\nAccount:\n\tName: Test Account\n\n"
    )


def test_str_method_with_categories():
    """Test the string representation of a Qif instance with categories"""
    qif = Qif()
    qif.add_category(Category(name="Test Category"))
    print(repr(str(qif)))
    assert str(qif) == (
        "QIF\n"
        "===\n\n"
        "Categories\n"
        "----------\n\n"
        "Category:\n\t"
        "Name: Test Category\n\t"
        "Category Type: expense\n\t"
        "Hierarchy: Test Category\n\n"
    )


def test_str_method_with_classes():
    """Test the string representation of a Qif instance with classes"""
    qif = Qif()
    qif.add_class(Class(name="Test Class"))
    assert str(qif) == (
        "QIF\n"
        "===\n\n"
        "Classes\n"
        "-------\n\n"
        "Class:\n\t"
        "Name: Test Class\n\t"
        "Categories: 0\n\n"
    )


def test_str_method_with_all():
    """Test the string representation of a Qif instance with all data"""
    qif = Qif()
    qif.add_account(Account(name="Test Account"))
    qif.add_class(Class(name="Test Class"))
    qif.add_category(Category(name="Test Category"))
    assert str(qif) == (
        "QIF\n"
        "===\n\n"
        "Accounts\n"
        "--------\n\n"
        "Account:\n\t"
        "Name: Test Account\n\n"
        "Categories\n"
        "----------\n\n"
        "Category:\n\t"
        "Name: Test Category\n\t"
        "Category Type: expense\n\t"
        "Hierarchy: Test Category\n\n"
        "Classes\n"
        "-------\n\n"
        "Class:\n\t"
        "Name: Test Class\n\t"
        "Categories: 0\n\n"
    )


def test_from_list():
    """Test that the from_list method raises a NotImplementedError"""
    qif = Qif()
    with pytest.raises(NotImplementedError):
        qif.from_list([])


def test_parse_string_works():
    """Test parsing a QIF file"""
    qif_string = (
        "!Type:Class\n"
        "NTest class\n"
        "DThis is just a class I added here for test purposes\n"
        "^\n"
        "!Type:Class\n"
        "NTest class 2\n"
        "DThis is just a class I added here for test purposes\n"
        "^\n"
    )
    Qif.parse_string(qif_string)


def test_parse_empty_string():
    """Test parsing an empty QIF file"""
    with pytest.raises(ParserException):
        Qif.parse("")


def test_parse_file_works(qif_file):
    """Test parsing a QIF file"""
    Qif.parse(qif_file)


def test_parse_file_works_with_oth_a_account(qif_file_with_oth_a_account):
    """Test parsing a QIF file with an OTH account"""
    Qif.parse(qif_file_with_oth_a_account)


def test_parse_empty_file(empty_file):
    """Test parsing an empty QIF file"""
    with pytest.raises(ParserException):
        Qif.parse(empty_file)


def test_parse_non_qif_file(txt_file):
    """Test parsing a file that is not a QIF file"""
    with pytest.raises(ParserException):
        Qif.parse(txt_file)


def test_parse_nonexistent_file():
    """Test parsing a file that does not exist"""
    with pytest.raises(ParserException):
        Qif.parse("nonexistent_file.qif")


def test_parsed_accounts(qif_file):
    """Test that the accounts are parsed correctly"""
    qif = Qif.parse(qif_file)

    # Validate the account is the default account
    assert len(qif.accounts) == 1
    account_name = "Quiffen Default Account"
    account = qif.accounts[account_name]

    assert account.name == account_name
    assert account.desc == (
        "The default account created by Quiffen when no other accounts were present"
    )


def test_parsed_transactions(qif_file):
    """Test that the transactions are parsed correctly"""
    qif = Qif.parse(qif_file)
    account = qif.accounts["Quiffen Default Account"]

    # Validate the transactions
    transactions = account.transactions
    assert len(transactions) == 2

    # Validate bank transactions
    bank_transactions = transactions[AccountType.BANK]
    assert len(bank_transactions) == 6

    # Validate the investment transactions
    investment_transactions = transactions[AccountType.INVST]
    assert len(investment_transactions) == 1


def test_parsed_categories(qif_file):
    """Test that the categories are parsed correctly"""
    qif = Qif.parse(qif_file)

    # Validate categories
    assert len(qif.categories) == 4
    expected_categories = ["Bills", "Food", "Investments", "Miscellaneous"]
    assert sorted(qif.categories.keys()) == expected_categories

    # Validate bills category
    bills = qif.categories["Bills"]
    assert len(bills.children) == 1
    assert bills.children[0].name == "Cell Phone"
    assert bills.children[0].parent == bills

    # Validate food category
    food = qif.categories["Food"]
    assert len(food.children) == 1
    assert food.children[0].name == "Groceries"
    assert food.children[0].parent == food

    # Validate investments category
    investments = qif.categories["Investments"]
    investment_child_categories = sorted(investments.children)
    assert len(investment_child_categories) == 2
    assert investment_child_categories[0].name == "Bonds"
    assert investment_child_categories[0].parent == investments
    assert investment_child_categories[1].name == "Stocks"
    assert investment_child_categories[1].parent == investments

    # Validate miscellaneous category
    miscellaneous = qif.categories["Miscellaneous"]
    assert len(miscellaneous.children) == 0


def test_parsed_classes(qif_file):
    """Test that the classes are parsed correctly"""
    qif = Qif.parse(qif_file)

    # Validate classes
    assert len(qif.classes) == 2
    assert "Test class" in qif.classes
    assert "Test class 2" in qif.classes

    # Validate test class
    test_class = qif.classes["Test class"]
    assert test_class.name == "Test class"
    assert test_class.desc == ("This is just a class I added here for test purposes")
    assert len(test_class.categories) == 1
    cell_phone = test_class.categories[0]
    assert cell_phone.name == "Cell Phone"
    assert cell_phone.parent is not None
    assert cell_phone.parent.name == "Bills"

    # Validate test class 2
    test_class_2 = qif.classes["Test class 2"]
    assert test_class_2.name == "Test class 2"
    assert test_class_2.desc == ("This is just a class I added here for test purposes")
    assert len(test_class_2.categories) == 0


def test_parsed_securities(qif_file):
    """Test that the securities are parsed correctly

    Relates to issue #30
    https://github.com/isaacharrisholt/quiffen/issues/30
    """
    qif = Qif.parse(qif_file)

    assert len(qif.securities) == 3
    assert sorted(qif.securities.keys()) == [
        "G002864",
        "M039728",
        "USD0000",
    ]

    assert qif.securities["G002864"].name == ""
    assert qif.securities["G002864"].symbol == "G002864"
    assert qif.securities["G002864"].type == "Stock"
    assert qif.securities["G002864"].goal == "Growth"

    assert qif.securities["M039728"].name == ""
    assert qif.securities["M039728"].symbol == "M039728"
    assert qif.securities["M039728"].type == "Stock"
    assert qif.securities["M039728"].goal == "Growth"

    assert qif.securities["USD0000"].name == ""
    assert qif.securities["USD0000"].symbol == "USD0000"
    assert qif.securities["USD0000"].type == "Stock"
    assert qif.securities["USD0000"].goal == "Growth"


def test_parse_nosymbol(qif_file_with_no_symbol):
    """Test that a security with no symbol can still be parsed without error"""
    qif = Qif.parse(qif_file_with_no_symbol)

    assert list(qif.securities.keys()) == ["Example"]


def test_add_account():
    """Test adding an account"""
    qif = Qif()
    account = Account(name="Test Account")
    qif.add_account(account)
    assert len(qif.accounts) == 1
    assert qif.accounts["Test Account"] == account


def test_add_existing_account():
    """Test adding an account that already exists"""
    qif = Qif()
    account = Account(name="Test Account")
    account2 = Account(name="Test Account")
    qif.add_account(account)
    qif.add_account(account2)
    assert len(qif.accounts) == 1
    assert "Test Account" in qif.accounts


def test_remove_account():
    """Test removing an account"""
    qif = Qif()
    account = Account(name="Test Account")
    qif.add_account(account)
    removed = qif.remove_account("Test Account")
    assert len(qif.accounts) == 0
    assert removed == account


def test_remove_nonexistent_account():
    """Test removing an account that does not exist"""
    qif = Qif()
    with pytest.raises(KeyError):
        qif.remove_account("Test Account")


def test_add_category():
    """Test adding a category"""
    qif = Qif()
    category = Category(name="Test Category")
    qif.add_category(category)
    assert len(qif.categories) == 1
    assert qif.categories["Test Category"] == category


def test_add_existing_category():
    """Test adding a category that already exists"""
    qif = Qif()
    category = Category(name="Test Category")
    category2 = Category(name="Test Category")
    qif.add_category(category)
    qif.add_category(category2)
    assert len(qif.categories) == 1
    assert "Test Category" in qif.categories


def test_remove_category():
    """Test removing a category"""
    qif = Qif()
    category = Category(name="Test Category")
    qif.add_category(category)
    removed = qif.remove_category("Test Category")
    assert len(qif.categories) == 0
    assert removed == category


def test_remove_category_keep_children():
    """Test removing a category and keeping its children"""
    qif = Qif()
    category = Category(name="Test Category")
    child_category = Category(name="Child Category", parent=category)
    qif.add_category(category)
    qif.add_category(child_category)
    removed = qif.remove_category("Test Category", keep_children=True)
    assert len(qif.categories) == 1
    assert "Child Category" in qif.categories
    assert removed == category


def test_remove_category_nonexistent_category():
    """Test removing a category that does not exist"""
    qif = Qif()
    with pytest.raises(KeyError):
        qif.remove_category("Test Category")


def test_add_class():
    """Test adding a class"""
    qif = Qif()
    cls = Class(name="Test Class")
    qif.add_class(cls)
    assert len(qif.classes) == 1
    assert qif.classes["Test Class"] == cls


def test_add_existing_class():
    """Test adding a class that already exists"""
    qif = Qif()
    cls = Class(name="Test Class")
    cls2 = Class(name="Test Class")
    qif.add_class(cls)
    qif.add_class(cls2)
    assert len(qif.classes) == 1
    assert "Test Class" in qif.classes


def test_remove_class():
    """Test removing a class"""
    qif = Qif()
    cls = Class(name="Test Class")
    qif.add_class(cls)
    removed = qif.remove_class("Test Class")
    assert len(qif.classes) == 0
    assert removed == cls


def test_remove_nonexistent_class():
    """Test removing a class that does not exist"""
    qif = Qif()
    with pytest.raises(KeyError):
        qif.remove_class("Test Class")


def test_add_security():
    """Test adding a security"""
    qif = Qif()
    security = Security(symbol="TEST")
    qif.add_security(security)
    assert len(qif.securities) == 1
    assert qif.securities["TEST"] == security


def test_add_existing_security():
    """Test adding a security that already exists"""
    qif = Qif()
    security = Security(symbol="TEST")
    security2 = Security(symbol="TEST")
    qif.add_security(security)
    qif.add_security(security2)
    assert len(qif.securities) == 1
    assert "TEST" in qif.securities


def test_remove_security():
    """Test removing a security"""
    qif = Qif()
    security = Security(symbol="TEST")
    qif.add_security(security)
    removed = qif.remove_security("TEST")
    assert len(qif.securities) == 0
    assert removed == security


def test_remove_nonexistent_security():
    """Test removing a security that does not exist"""
    qif = Qif()
    with pytest.raises(KeyError):
        qif.remove_security("TEST")


def test_to_qif(qif_file):
    """Test the to_qif method

    The most important thing to test here is that the output contains all the
    data that was parsed from the input file.

    As such, the test file is parsed and then the output is compared to the
    original file.
    """
    qif = Qif.parse(qif_file)
    test_file = qif_file.parent / "test_output.qif"
    qif.to_qif(test_file)

    resulting_qif = Qif.parse(test_file)
    assert qif.accounts == resulting_qif.accounts
    assert sorted(qif.categories) == sorted(resulting_qif.categories)
    assert qif.classes == resulting_qif.classes
    test_file.unlink()


# pylint: disable=protected-access
def test_get_data_dicts_transactions():
    """Test the get_data_dicts method with transactions"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)
    qif.add_account(account)
    data_dicts = qif._get_data_dicts(data_type=QifDataType.TRANSACTIONS)

    assert len(data_dicts) == 1
    expected = transaction.to_dict()
    expected["date"] = "2019-01-01"  # Dates are converted to strings
    assert data_dicts[0] == expected


def test_get_data_dicts_transactions_with_date_format_and_ignore():
    """Test the get_data_dicts method with transactions and date format and
    ignore"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)
    qif.add_account(account)
    data_dicts = qif._get_data_dicts(
        data_type=QifDataType.TRANSACTIONS,
        date_format="%d/%m/%Y",
        ignore=["payee", "memo"],
    )

    assert len(data_dicts) == 1
    expected = transaction.to_dict()
    expected["date"] = "01/01/2019"  # Dates are converted to strings
    del expected["payee"]
    del expected["memo"]
    assert data_dicts[0] == expected


def test_get_data_dicts_categories():
    """Test the get_data_dicts method with categories"""
    qif = Qif()
    category = Category(name="Test Category")
    qif.add_category(category)
    data_dicts = qif._get_data_dicts(data_type=QifDataType.CATEGORIES)

    assert len(data_dicts) == 1
    assert data_dicts[0] == category.to_dict()


def test_get_data_dicts_classes():
    """Test the get_data_dicts method with classes"""
    qif = Qif()
    cls = Class(name="Test Class")
    qif.add_class(cls)
    data_dicts = qif._get_data_dicts(data_type=QifDataType.CLASSES)

    assert len(data_dicts) == 1
    assert data_dicts[0] == cls.to_dict()


def test_get_data_dicts_accounts():
    """Test the get_data_dicts method with accounts"""
    qif = Qif()
    account = Account(name="Test Account")
    account.set_header(AccountType.BANK)
    qif.add_account(account)
    data_dicts = qif._get_data_dicts(data_type=QifDataType.ACCOUNTS)

    assert len(data_dicts) == 1
    expected = account.to_dict()
    expected.pop("_last_header", None)
    assert data_dicts[0] == expected


def test_get_data_dicts_investments():
    """Test the get_data_dicts method with investments"""
    qif = Qif()
    account = Account(name="Test Account")
    investment = Investment(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        security="Test Security",
        price=Decimal("10"),
    )
    account.set_header(AccountType.INVST)
    account.add_transaction(investment)
    qif.add_account(account)
    data_dicts = qif._get_data_dicts(data_type=QifDataType.INVESTMENTS)

    assert len(data_dicts) == 1
    expected = investment.to_dict()
    expected["date"] = "2019-01-01"  # Dates are converted to strings
    assert data_dicts[0] == expected


def test_get_data_dicts_securities():
    """Test the get_data_dicts method with securities"""
    qif = Qif()
    security = Security(symbol="TEST")
    qif.add_security(security)
    data_dicts = qif._get_data_dicts(data_type=QifDataType.SECURITIES)

    assert len(data_dicts) == 1
    assert data_dicts[0] == security.to_dict()


# pylint: enable=protected-access


def test_to_csv_transactions():
    """Test the to_csv method with transactions"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)
    qif.add_account(account)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(path=csv_file, data_type=QifDataType.TRANSACTIONS)

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)
        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 20

        results = list(reader)
        assert len(results) == 1
        assert results[0]["date"] == "2019-01-01"
        assert results[0]["amount"] == "100"
        assert results[0]["payee"] == "Test Payee"
        assert results[0]["memo"] == "Test Memo"
        assert "Test Category" in results[0]["category"]
    csv_file.unlink()


def test_to_csv_categories():
    """Test the to_csv method with categories"""
    qif = Qif()
    category = Category(name="Test Category")
    qif.add_category(category)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(path=csv_file, data_type=QifDataType.CATEGORIES)

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)

        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 9

        results = list(reader)
        assert len(results) == 1
        assert results[0]["name"] == "Test Category"
    csv_file.unlink()


def test_to_csv_classes():
    """Test the to_csv method with classes"""
    qif = Qif()
    cls = Class(name="Test Class", desc="Test Description")
    qif.add_class(cls)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(path=csv_file, data_type=QifDataType.CLASSES)

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)

        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 3

        results = list(reader)
        assert len(results) == 1
        assert results[0]["name"] == "Test Class"
        assert results[0]["desc"] == "Test Description"
    csv_file.unlink()


def test_to_csv_accounts():
    """Test the to_csv method with accounts"""
    qif = Qif()
    account = Account(name="Test Account", account_type=AccountType.BANK)
    account.set_header(AccountType.BANK)
    qif.add_account(account)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(path=csv_file, data_type=QifDataType.ACCOUNTS)

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)

        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 7

        results = list(reader)
        assert len(results) == 1
        assert results[0]["name"] == "Test Account"
        assert results[0]["account_type"] == "Bank"
    csv_file.unlink()


def test_to_csv_investments():
    """Test the to_csv method with investments"""
    qif = Qif()
    account = Account(name="Test Account")
    investment = Investment(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        security="Test Security",
        price=Decimal("10"),
    )
    account.set_header(AccountType.INVST)
    account.add_transaction(investment)
    qif.add_account(account)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(path=csv_file, data_type=QifDataType.INVESTMENTS)

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)

        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 13

        results = list(reader)
        assert len(results) == 1
        assert results[0]["date"] == "2019-01-01"
        assert results[0]["amount"] == "100"
        assert results[0]["security"] == "Test Security"
        assert results[0]["price"] == "10"
    csv_file.unlink()


def test_to_csv_securities():
    """Test the to_csv method with securities"""
    qif = Qif()
    security = Security(symbol="TEST")
    qif.add_security(security)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(path=csv_file, data_type=QifDataType.SECURITIES)

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)

        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 5

        results = list(reader)
        assert len(results) == 1
        assert results[0]["symbol"] == "TEST"
    csv_file.unlink()


def test_to_csv_transactions_with_date_format_and_ignore_list():
    """Test the to_csv method with transactions with date format and ignore"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 2, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)
    qif.add_account(account)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(
        path=csv_file,
        data_type=QifDataType.TRANSACTIONS,
        date_format="%m/%d/%Y",
        ignore=["payee"],
    )

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)

        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 19

        results = list(reader)
        assert len(results) == 1
        assert results[0]["date"] == "02/01/2019"
        assert results[0]["amount"] == "100"
        assert "payee" not in results[0]
        assert results[0]["memo"] == "Test Memo"
        assert "Test Category" in results[0]["category"]
    csv_file.unlink()


def test_to_csv_transactions_multiple():
    """Test the to_csv method with multiple transactions"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)

    transaction2 = transaction.model_copy()
    transaction2.amount = Decimal("200")

    account.add_transaction(transaction2)
    qif.add_account(account)

    csv_file = Path(__file__).parent / "test_files" / "test_output.csv"
    qif.to_csv(path=csv_file, data_type=QifDataType.TRANSACTIONS)

    assert csv_file.exists()
    with csv_file.open() as f:
        reader = csv.DictReader(f)

        assert reader.fieldnames is not None
        assert len(reader.fieldnames) == 20

        results = list(reader)
        assert len(results) == 2
        assert results[0]["date"] == "2019-01-01"
        assert results[0]["amount"] == "100"
        assert results[0]["payee"] == "Test Payee"
        assert results[0]["memo"] == "Test Memo"
        assert "Test Category" in results[0]["category"]
        assert results[1]["date"] == "2019-01-01"
        assert results[1]["amount"] == "200"
        assert results[1]["payee"] == "Test Payee"
        assert results[1]["memo"] == "Test Memo"
        assert "Test Category" in results[1]["category"]
    csv_file.unlink()


def test_to_dataframe_transactions():
    """Test the to_dataframe method with transactions"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)
    qif.add_account(account)

    df = qif.to_dataframe(data_type=QifDataType.TRANSACTIONS)

    assert df.shape == (1, 20)
    assert df["date"][0] == datetime(2019, 1, 1)
    assert df["amount"][0] == 100
    assert df["payee"][0] == "Test Payee"
    assert df["memo"][0] == "Test Memo"
    assert df["category"][0]["name"] == "Test Category"


def test_to_dataframe_categories():
    """Test the to_dataframe method with categories"""
    qif = Qif()
    category = Category(name="Test Category")
    qif.add_category(category)

    df = qif.to_dataframe(data_type=QifDataType.CATEGORIES)

    assert df.shape == (1, 9)
    assert df["name"][0] == "Test Category"
    assert df["parent"][0] is None


def test_to_dataframe_classes():
    """Test the to_dataframe method with classes"""
    qif = Qif()
    cls = Class(name="Test Class", desc="Test Description")
    qif.add_class(cls)

    df = qif.to_dataframe(data_type=QifDataType.CLASSES)

    assert df.shape == (1, 3)
    assert df["name"][0] == "Test Class"
    assert df["desc"][0] == "Test Description"


def test_to_dataframe_securities():
    """Test the to_dataframe method with securities"""
    qif = Qif()
    security = Security(symbol="TEST")
    qif.add_security(security)

    df = qif.to_dataframe(data_type=QifDataType.SECURITIES)

    assert df.shape == (1, 5)
    assert df["symbol"][0] == "TEST"
    assert df["name"][0] is None
    assert df["type"][0] is None
    assert df["goal"][0] is None


def test_to_dataframe_accounts():
    """Test the to_dataframe method with accounts"""
    qif = Qif()
    account = Account(name="Test Account", account_type=AccountType.BANK)
    qif.add_account(account)

    df = qif.to_dataframe(data_type=QifDataType.ACCOUNTS)

    assert df.shape == (1, 7)
    assert df["name"][0] == "Test Account"
    assert df["account_type"][0] == "Bank"


def test_to_dataframe_investments():
    """Test the to_dataframe method with investments"""
    qif = Qif()
    account = Account(name="Test Account")
    investment = Investment(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        security="Test Security",
        price=Decimal("10"),
    )
    account.set_header(AccountType.INVST)
    account.add_transaction(investment)
    qif.add_account(account)

    df = qif.to_dataframe(data_type=QifDataType.INVESTMENTS)

    assert df.shape == (1, 13)
    assert df["date"][0] == datetime(2019, 1, 1)
    assert df["amount"][0] == 100
    assert df["security"][0] == "Test Security"
    assert df["price"][0] == 10


def test_to_dataframe_transactions_with_ignore_list():
    """Test the to_dataframe method with transactions and ignore"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 2, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)
    qif.add_account(account)

    df = qif.to_dataframe(
        data_type=QifDataType.TRANSACTIONS,
        ignore=["payee"],
    )

    assert df.shape == (1, 19)
    assert df["date"][0] == datetime(2019, 2, 1)
    assert df["amount"][0] == 100
    assert "payee" not in df.columns
    assert df["memo"][0] == "Test Memo"
    assert df["category"][0]["name"] == "Test Category"


def test_to_dataframe_transactions_multiple():
    """Test the to_dataframe method with multiple transactions"""
    qif = Qif()
    account = Account(name="Test Account")
    transaction = Transaction(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
        payee="Test Payee",
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    account.set_header(AccountType.BANK)
    account.add_transaction(transaction)

    transaction2 = transaction.model_copy()
    transaction2.amount = Decimal("200")

    account.add_transaction(transaction2)
    qif.add_account(account)

    df = qif.to_dataframe(data_type=QifDataType.TRANSACTIONS)

    assert df.shape == (2, 20)
    assert df["date"][0] == datetime(2019, 1, 1)
    assert df["amount"][0] == 100
    assert df["payee"][0] == "Test Payee"
    assert df["memo"][0] == "Test Memo"
    assert df["category"][0]["name"] == "Test Category"
    assert df["date"][1] == datetime(2019, 1, 1)
    assert df["amount"][1] == 200
    assert df["payee"][1] == "Test Payee"
    assert df["memo"][1] == "Test Memo"
    assert df["category"][1]["name"] == "Test Category"


def test_transaction_before_account_definition_1(qif_file):
    """Tests that a transaction can be parsed before the account
    definition.

    Relates to pull #32.
    https://github.com/isaacharrisholt/quiffen/pull/32
    """
    test_file = qif_file.parent / "test_transaction_before_account_definition_1.qif"
    qif = Qif.parse(test_file)
    acc = qif.accounts["Quiffen Default Account"]

    assert len(acc.transactions) == 1
    assert len(acc.transactions[AccountType.BANK]) == 2


def test_transaction_before_account_definition_2(qif_file):
    """Tests that a transaction can be parsed before the account
    definition.

    Relates to pull #32.
    """
    test_file = qif_file.parent / "test_transaction_before_account_definition_2.qif"
    qif = Qif.parse(test_file)
    acc = qif.accounts["Quiffen Default Account"]

    assert len(acc.transactions) == 1
    assert len(acc.transactions[AccountType.BANK]) == 2


def test_empty_qif():
    qif = Qif()
    assert qif.to_qif() == ""


def test_transaction_account_type_qif():
    qif = Qif()
    account = Account(name="Test Account")
    qif.add_account(account)
    transaction = Transaction(
        date=datetime(2019, 1, 1),
        amount=Decimal("100"),
    )
    account.set_header(AccountType.CASH)
    account.add_transaction(transaction)
    assert (
        qif.to_qif()
        == """!Account
NTest Account
^
!Type:Cash
D2019-01-01
T100
^
"""
    )


def test_clear_autoswitch_ignored(qif_file_with_clear_autoswitch):
    """Tests that `!Clear:AutoSwitch` flag exported by Quicken and GNUCash
    is ignored.

    Relates to discussion #89.
    """
    qif = Qif.parse(qif_file_with_clear_autoswitch)
    assert len(qif.accounts) == 1
    assert list(qif.accounts.keys()) == ["My Bank Account"]


def test_option_autoswitch_ignored(qif_file_with_option_autoswitch):
    """Tests that `!Opiton:AutoSwitch` flag exported by Quicken
    is ignored.

    Relates to discussion #92.
    """
    qif = Qif.parse(qif_file_with_option_autoswitch)
    assert len(qif.accounts) == 1
    assert list(qif.accounts.keys()) == ["My Bank Account"]


def test_split_to_account(qif_file_with_split_to_account):
    """Tests that a split to an account is recorded appropriately.

    Relates to discussion #94
    """
    qif = Qif.parse(qif_file_with_split_to_account)
    account = qif.accounts["Quiffen Default Account"]

    # Validate the splits
    transactions = account.transactions
    bank_transactions = transactions[AccountType.BANK]
    split_transaction = bank_transactions[0]

    assert split_transaction.amount == Decimal("-10")
    assert isinstance(split_transaction, Transaction)
    assert split_transaction.splits[0].to_account == "An Account"
    assert split_transaction.splits[0].category is None
    assert split_transaction.splits[1].to_account is None
    assert split_transaction.splits[1].category == Category(name="A Category")


def test_unknown_account_type(qif_file_with_unknown_account_type):
    """Tests that unknown account types are handled.

    Relates to discussion #95.
    """
    qif = Qif.parse(qif_file_with_unknown_account_type)
    assert qif.accounts["Portfolio Account"].account_type == AccountType.UNKNOWN
    # Oddly, the transactions are grouped under a known account type.
    transactions = qif.accounts["Portfolio Account"].transactions
    assert AccountType.INVST in transactions


def test_to_csv_circular_reference(qif_file_circular_reference):
    """Tests that a circular reference is handled.

    Relates to discussion #113.
    """
    qif = Qif.parse(qif_file_circular_reference, day_first=False)
    qif.to_csv()



from datetime import datetime
from decimal import Decimal

import pytest

from quiffen import Category, Class, Split, Transaction


def test_create_transaction():
    """Test creating a transaction"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(100),
    )
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 100
    assert transaction.check_number is None
    assert transaction.payee is None
    assert transaction.memo is None
    assert transaction.payee_address is None
    assert transaction.category is None
    assert not transaction.splits


def test_create_transaction_more_fields():
    """Test creating a transaction with more fields"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(100),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
    )
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 100
    assert transaction.check_number == 1
    assert transaction.payee == "Test Payee"
    assert transaction.memo == "Test Memo"
    assert transaction.payee_address == "Test Address"
    assert transaction.category is None
    assert not transaction.splits


def test_create_transaction_with_splits():
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(100),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100))],
    )
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == Decimal(100)
    assert transaction.check_number == 1
    assert transaction.payee == "Test Payee"
    assert transaction.memo == "Test Memo"
    assert transaction.payee_address == "Test Address"
    assert transaction.category == Category(name="Test Category")
    assert transaction.splits == [Split(amount=Decimal(100))]


def test_create_transaction_with_splits_percent_too_high():
    """Test creating a transaction with splits where the percent is too high"""
    with pytest.raises(ValueError):
        Transaction(
            date=datetime(2022, 2, 1),
            amount=Decimal(100),
            splits=[
                Split(amount=Decimal(100), percent=Decimal(50)),
                Split(amount=Decimal(200), percent=Decimal(60)),
            ],
        )


def test_create_transaction_with_splits_exactly_100_percent():
    """Test creating a transaction with splits where the percent is exactly 100.

    Relates to issue #39.
    https://github.com/isaacharrisholt/quiffen/issues/39
    """
    transaction_list = [
        "D12/1/22",
        "PPayee",
        "T5,382.39",
        "NDEP",
        "LSplit",
        "SCategory A",
        "$-120.83",
        "SCategory B",
        "$-2,100.96",
        "SCategory C",
        "$-729.15",
        "SCategory D",
        "$8,333.33",
    ]
    # Should not raise an error
    Transaction.from_list(transaction_list)


def test_create_transaction_with_splits_percent_rounding_error_less_than_01():
    """
    Test creating a transaction with splits amounts that could lead to a sum
    of percentages higher than 100.01 when they are rounded at two decimal
    positions.

    With the data below, the sum of the percentages rounded to 2 decimal positions
        gives 100.02

    Relates to issue #63.
    https://github.com/isaacharrisholt/quiffen/issues/63
    """
    transaction_list = [
        "D13/05'2009",
        "CX",
        "T-1,910.22",
        "PPayee",
        "LMetaCategory A",
        "SMetaCategory A",
        "EMetaCategory A",
        "$-29.23",
        "SCategory B",
        "ECategory B",
        "$-697.91",
        "SCategory B",
        "ECategory B",
        "$-34.37",
        "SCategory B",
        "ECategory B",
        "$-22.47",
        "SCategory B",
        "ECategory B",
        "$-44.06",
        "SCategory B",
        "ECategory B",
        "$-18.00",
        "SCategory B",
        "ECategory B",
        "$-58.00",
        "SCategory B",
        "ECategory B",
        "$-4.95",
        "SCategory B",
        "ECategory B",
        "$-18.65",
        "SCategory B",
        "ECategory B",
        "$-53.68",
        "SCategory B",
        "ECategory B",
        "$-13.55",
        "SCategory B",
        "ECategory B",
        "$-7.58",
        "SCategory B",
        "ECategory B",
        "$-140.00",
        "SCategory B",
        "ECategory B",
        "$-661.52",
        "SCategory B",
        "ECategory B",
        "$-106.25",
    ]
    # Should not raise an error
    Transaction.from_list(transaction_list)


def test_create_transaction_with_splits_amount_too_high():
    """Test creating a transaction with splits where the amount is too high"""
    with pytest.raises(ValueError):
        Transaction(
            date=datetime(2022, 2, 1),
            amount=Decimal(100),
            splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
        )


def test_eq_success():
    """Test equality"""
    transaction1 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
    )
    transaction2 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
    )
    assert transaction1 == transaction2

    transaction3 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    transaction4 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    assert transaction3 == transaction4


def test_eq_failure():
    """Test inequality"""
    transaction1 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
    )
    transaction2 = Transaction(
        date=datetime(2022, 2, 2),
        amount=Decimal(1000),
    )
    assert transaction1 != transaction2

    transaction3 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    transaction4 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[
            Split(amount=Decimal(100)),
            Split(amount=Decimal(200)),
            Split(amount=Decimal(300)),
        ],
    )
    assert transaction3 != transaction4

    transaction5 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    transaction5.splits[0].amount = Decimal(300)
    assert transaction3 != transaction5

    transaction6 = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo 2",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    assert transaction3 != transaction6


def test_str_method():
    """Test the string method"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    assert str(transaction) == (
        "Transaction:\n\t"
        "Date: 2022-02-01 00:00:00\n\t"
        "Amount: 1000\n\t"
        "Memo: Test Memo\n\t"
        "Payee: Test Payee\n\t"
        "Payee Address: Test Address\n\t"
        "Category: Test Category\n\t"
        "Check Number: 1\n\t"
        "Splits: 2"
    )


def test_is_split():
    """Test the is_split property"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    assert transaction.is_split

    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
    )
    assert not transaction.is_split


def test_add_split():
    """Test the add_split method"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
    )
    transaction.add_split(Split(amount=Decimal(100)))
    assert transaction.splits == [Split(amount=Decimal(100))]
    transaction.add_split(Split(amount=Decimal(200)))
    assert transaction.splits == [
        Split(amount=Decimal(100)),
        Split(amount=Decimal(200)),
    ]


def test_add_split_percent_too_high():
    """Test the add_split method where the sum of the percentages is too high"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100), percent=Decimal(60))],
    )
    with pytest.raises(ValueError):
        transaction.add_split(Split(amount=Decimal(100), percent=Decimal(60)))


def test_add_split_amount_too_high():
    """Test the add_split method where the sum of the amounts is too high"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100))],
    )
    with pytest.raises(ValueError):
        transaction.add_split(Split(amount=Decimal(950)))


def test_add_splits_exactly_100_percent():
    """Test adding splits that add up to 100 percent

    Relates to issue #39.
    https://github.com/isaacharrisholt/quiffen/issues/39
    """
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal("5382.39"),
    )

    splits = [
        Split(amount=Decimal("-121.83")),
        Split(amount=Decimal("-2101.96")),
        Split(amount=Decimal("-730.15")),
        Split(amount=Decimal("8332.33")),
    ]

    for split in splits:
        # Should not raise an error
        transaction.add_split(split)


def test_remove_splits_one_filter():
    """Test the remove_splits method with one filter"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[
            Split(amount=Decimal(100)),
            Split(amount=Decimal(200), memo="Test Memo"),
        ],
    )
    transaction.remove_splits(memo="Test Memo")
    assert transaction.splits == [Split(amount=Decimal(100))]


def test_remove_splits_multiple_filters():
    """Test the remove_splits method with multiple filters"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[
            Split(amount=Decimal(100)),
            Split(amount=Decimal(100), memo="Test Memo"),
        ],
    )
    transaction.remove_splits(memo="Test Memo", amount=Decimal(100))
    assert transaction.splits == [Split(amount=Decimal(100))]


def test_remove_splits_no_match():
    """Test the remove_splits method with no match"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[
            Split(amount=Decimal(100)),
            Split(amount=Decimal(100), memo="Test Memo"),
        ],
    )
    transaction.remove_splits(memo="Test Memo", amount=Decimal(100), check_number=1)
    assert transaction.splits == [
        Split(amount=Decimal(100)),
        Split(amount=Decimal(100), memo="Test Memo"),
    ]


def test_remove_splits_no_filters():
    """Test the remove_splits method with no filters"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[Split(amount=Decimal(100)), Split(amount=Decimal(200))],
    )
    transaction.remove_splits()
    assert not transaction.splits
    assert not transaction.is_split


def test_to_qif_no_splits_no_classes():
    """Test the to_qif method with no splits and no classes"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
    )
    print(repr(transaction.to_qif()))
    assert transaction.to_qif() == (
        "D2022-02-01\n"
        "T1000\n"
        "MTest Memo\n"
        "PTest Payee\n"
        "ATest Address\n"
        "LTest Category\n"
        "N1\n"
    )


def test_to_qif_no_split_with_class():
    """Test the to_qif method with no splits and a class"""
    parent = Category(name="Test Parent")
    child = Category(name="Test Child")
    parent.add_child(child)

    cls = Class(name="Test Class")
    cls.add_category(parent)

    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=child,
    )
    assert transaction.to_qif(classes={"Test Class": cls}) == (
        "D2022-02-01\n"
        "T1000\n"
        "MTest Memo\n"
        "PTest Payee\n"
        "ATest Address\n"
        "LTest Parent:Test Child/Test Class\n"
        "N1\n"
    )


def test_to_qif_with_splits_no_classes():
    """Test the to_qif method with splits and no classes"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=Category(name="Test Category"),
        splits=[
            Split(
                amount=Decimal(100),
                memo="Test Memo",
                category=Category(name="Test Split Category"),
            ),
            Split(amount=Decimal(200), memo="Test Memo 2"),
        ],
    )
    assert transaction.to_qif() == (
        "D2022-02-01\n"
        "T1000\n"
        "MTest Memo\n"
        "PTest Payee\n"
        "ATest Address\n"
        "LTest Category\n"
        "N1\n"
        "STest Split Category\n"
        "$100\n"
        "ETest Memo\n"
        "S\n"
        "$200\n"
        "ETest Memo 2\n"
    )


def test_to_qif_with_splits_with_classes():
    """Test the to_qif method with splits and classes"""
    parent = Category(name="Test Parent")
    child = Category(name="Test Child")
    parent.add_child(child)

    cls = Class(name="Test Class")
    cls.add_category(parent)

    split_category = Category(name="Test Split Category")
    cls.add_category(split_category)

    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        check_number=1,
        payee="Test Payee",
        memo="Test Memo",
        payee_address="Test Address",
        category=child,
        splits=[
            Split(
                amount=Decimal(100),
                memo="Test Memo",
                category=Category(name="Test Split Category"),
            ),
            Split(amount=Decimal(200), memo="Test Memo 2"),
        ],
    )
    assert transaction.to_qif(classes={"Test Class": cls}) == (
        "D2022-02-01\n"
        "T1000\n"
        "MTest Memo\n"
        "PTest Payee\n"
        "ATest Address\n"
        "LTest Parent:Test Child/Test Class\n"
        "N1\n"
        "STest Split Category/Test Class\n"
        "$100\n"
        "ETest Memo\n"
        "S\n"
        "$200\n"
        "ETest Memo 2\n"
    )


def test_from_list_no_splits_no_classes():
    """Test creating a transaction from a list of QIF strings"""
    qif_list = [
        "D2022-02-01",
        "T1000",
        "MTest Memo",
        "CTest Cleared",
        "PTest Payee",
        "ATest Address",
        "L[Test To Account]",  # Brackets denote to account
        "LTest Category",  # No brackets denote category
        "N1",
        "FFalse",
        "12022-03-01",  # First payment date
        "234",  # Loan length
        "312",  # Number of payments
        "42",  # Periods per year
        "51.23",  # Interest rate
        "61000",  # Current loan balance
        "710000",  # Original loan amount
    ]
    transaction, _ = Transaction.from_list(qif_list)
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 1000
    assert transaction.memo == "Test Memo"
    assert transaction.cleared == "Test Cleared"
    assert transaction.payee == "Test Payee"
    assert transaction.payee_address == "Test Address"
    assert transaction.to_account == "Test To Account"
    assert transaction.category == Category(name="Test Category")
    assert transaction.check_number == 1
    assert not transaction.is_split
    assert transaction.first_payment_date == datetime(2022, 3, 1)
    assert transaction.loan_length == 34
    assert transaction.num_payments == 12
    assert transaction.periods_per_annum == 2
    assert transaction.interest_rate == Decimal("1.23")
    assert transaction.current_loan_balance == 1000
    assert transaction.original_loan_amount == 10000


def test_from_list_no_split_with_class():
    """Test creating a transaction from a list of QIF strings with no splits but
    that does define a QIF class"""
    qif_list = [
        "D2022-02-01",
        "T1000",
        "L[Test To Account]",  # Brackets denote to account
        "LTest Category/Class Name",  # / denotes a class after the category
    ]
    transaction, classes = Transaction.from_list(qif_list)
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 1000
    assert transaction.to_account == "Test To Account"
    assert transaction.category == Category(name="Test Category")

    assert classes == {
        "Class Name": Class(
            name="Class Name",
            categories=[Category(name="Test Category")],
        ),
    }


def test_from_list_with_splits_no_classes():
    """Test creating a transaction from a list of QIF strings with splits"""
    qif_list = [
        "D2022-02-01",
        "T1000",
        "L[Test To Account]",  # Brackets denote to account
        "LTest Category",  # No brackets denote category
        "STest Split Category 1",
        "ETest Split Memo",
        "$100",
        "STest Split Category 2",
        "EMemo",
        "$100",
        "%10",
    ]
    transaction, _ = Transaction.from_list(qif_list)
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 1000
    assert transaction.to_account == "Test To Account"
    assert transaction.category == Category(name="Test Category")
    assert transaction.is_split
    assert transaction.splits == [
        Split(
            category=Category(name="Test Split Category 1"),
            memo="Test Split Memo",
            amount=Decimal(100),
            percent=Decimal("10"),
        ),
        Split(
            category=Category(name="Test Split Category 2"),
            memo="Memo",
            amount=Decimal(100),
            percent=Decimal("10"),
        ),
    ]


def test_from_list_with_splits_with_classes():
    """Test creating a transaction from a list of QIF strings with splits and
    classes"""
    qif_list = [
        "D2022-02-01",
        "T1000",
        "L[Test To Account]",  # Brackets denote to account
        "LTest Category",  # No brackets denote category
        "STest Split Category 1/Class Name",
        "ETest Split Memo",
        "T100",
        "STest Split Category 2/Class Name",
        "EMemo",
        "$100",
        "%10",
    ]
    transaction, classes = Transaction.from_list(qif_list)
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 1000
    assert transaction.to_account == "Test To Account"
    assert transaction.category == Category(name="Test Category")
    assert transaction.is_split
    assert transaction.splits == [
        Split(
            category=Category(name="Test Split Category 1"),
            memo="Test Split Memo",
            amount=Decimal(100),
            percent=Decimal("10"),
        ),
        Split(
            category=Category(name="Test Split Category 2"),
            memo="Memo",
            amount=Decimal(100),
            percent=Decimal("10"),
        ),
    ]

    assert classes == {
        "Class Name": Class(
            name="Class Name",
            categories=[
                Category(name="Test Split Category 1"),
                Category(name="Test Split Category 2"),
            ],
        ),
    }


def test_from_list_multiple_categories():
    """Test creating a transaction from a list of QIF strings with multiple
    categories"""
    qif_list = [
        "D2022-02-01",
        "T1000",
        "L[Test To Account]",  # Brackets denote to account
        "LTest Category 1",
        "LTest Category 2",
    ]
    transaction, _ = Transaction.from_list(qif_list)
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 1000
    assert transaction.to_account == "Test To Account"
    assert transaction.category is not None
    assert transaction.category.name == "Test Category 2"
    assert transaction.category.parent is not None
    assert transaction.category.parent.name == "Test Category 1"


def test_from_string_default_separator():
    """Test creating a transaction from a string with the default separator"""
    qif_string = (
        "D2022-02-01\n"
        "T1000\n"
        "L[Test To Account]\n"  # Brackets denote to account
        "LTest Category"
    )
    transaction, _ = Transaction.from_string(qif_string)
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 1000
    assert transaction.to_account == "Test To Account"
    assert transaction.category == Category(name="Test Category")


def test_from_string_custom_separator():
    """Test creating a transaction from a string with a custom separator"""
    qif_string = (
        "D2022-02-01---"
        "T1000---"
        "L[Test To Account]---"  # Brackets denote to account
        "LTest Category"
    )
    transaction, _ = Transaction.from_string(qif_string, separator="---")
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 1000
    assert transaction.to_account == "Test To Account"
    assert transaction.category == Category(name="Test Category")


def test_to_dict():
    """Test converting a category to a dict"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        to_account="Test To Account",
        category=Category(name="Test Category"),
    )
    assert transaction.to_dict() == {
        "date": datetime(2022, 2, 1, 0, 0),
        "amount": Decimal("1000"),
        "memo": None,
        "cleared": None,
        "payee": None,
        "payee_address": None,
        "category": {
            "name": "Test Category",
            "desc": None,
            "tax_related": None,
            "category_type": "expense",
            "budget_amount": None,
            "tax_schedule_info": None,
            "hierarchy": "Test Category",
            "children": [],
            "parent": None,
        },
        "check_number": None,
        "reimbursable_expense": None,
        "small_business_expense": None,
        "to_account": "Test To Account",
        "first_payment_date": None,
        "loan_length": None,
        "num_payments": None,
        "periods_per_annum": None,
        "interest_rate": None,
        "current_loan_balance": None,
        "original_loan_amount": None,
        "line_number": None,
        "splits": [],
    }


def test_to_dict_with_ignore():
    """Test converting a category to a dict with ignored fields"""
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        to_account="Test To Account",
        category=Category(name="Test Category"),
    )
    assert transaction.to_dict(ignore=["to_account"]) == {
        "date": datetime(2022, 2, 1, 0, 0),
        "amount": Decimal("1000"),
        "memo": None,
        "cleared": None,
        "payee": None,
        "payee_address": None,
        "category": {
            "name": "Test Category",
            "desc": None,
            "tax_related": None,
            "category_type": "expense",
            "budget_amount": None,
            "tax_schedule_info": None,
            "hierarchy": "Test Category",
            "children": [],
            "parent": None,
        },
        "check_number": None,
        "reimbursable_expense": None,
        "small_business_expense": None,
        "first_payment_date": None,
        "loan_length": None,
        "num_payments": None,
        "periods_per_annum": None,
        "interest_rate": None,
        "current_loan_balance": None,
        "original_loan_amount": None,
        "line_number": None,
        "splits": [],
    }


def test_from_list_zero_value_with_splits():
    """Test creating a zero-value transaction from a list of QIF strings with
    splits

    Relates to issue #31.
    https://github.com/isaacharrisholt/quiffen/issues/31
    """
    qif_list = [
        "D2022-02-01",
        "T0",
        "L[Test To Account]",  # Brackets denote to account
        "LTest Category",  # No brackets denote category
        "STest Split Category 1",
        "ETest Split Memo",
        "T0",
        "STest Split Category 2",
        "EMemo",
        "$0",
    ]
    transaction, _ = Transaction.from_list(qif_list)
    assert transaction.date == datetime(2022, 2, 1)
    assert transaction.amount == 0
    assert transaction.to_account == "Test To Account"
    assert transaction.category == Category(name="Test Category")
    assert transaction.is_split
    assert transaction.splits == [
        Split(
            category=Category(name="Test Split Category 1"),
            memo="Test Split Memo",
            amount=Decimal(0),
            percent=None,
        ),
        Split(
            category=Category(name="Test Split Category 2"),
            memo="Memo",
            amount=Decimal(0),
            percent=None,
        ),
    ]


def test_check_number_allows_strings():
    """Test that the check number field allows strings

    Relates to issue #28.
    https://github.com/isaacharrisholt/quiffen/issues/28
    """
    transaction = Transaction(
        date=datetime(2022, 2, 1),
        amount=Decimal(1000),
        to_account="Test To Account",
        category=Category(name="Test Category"),
        check_number="Transfer",
    )
    assert transaction.check_number == "Transfer"


def test_from_list_invalid_characters_in_amount_field():
    """Test that the amount field correctly parses out invalid
    characters (non-numeric).

    Relates to issue #56.
    https://github.com/isaacharrisholt/quiffen/issues/56
    """
    qif_list = [
        "D2022-02-01",
        "T$1,0ab£c00",
        "L[Test To Account]",  # Brackets denote to account
        "LTest Category 1",
        "LTest Category 2",
    ]
    transaction, _ = Transaction.from_list(qif_list)
    assert transaction.amount == 1000



# pylint: disable=too-few-public-methods,protected-access,abstract-method


from quiffen.core.base import BaseModel, Field


def test_create_field():
    """Test creating a field"""
    field = Field(line_code="T", attr="test", type=str)
    assert field.line_code == "T"
    assert field.attr == "test"
    assert field.type is str


def test_sorting_fields():
    """Test sorting fields by line code"""
    field1 = Field(line_code="A", attr="a", type=str)
    field2 = Field(line_code="B", attr="b", type=str)
    field3 = Field(line_code="AA", attr="aa", type=str)

    assert sorted([field1, field2, field3]) == [field1, field3, field2]


def test_base_model_allows_extra_fields():
    """Test that the BaseModel allows extra fields"""

    class TestModel(BaseModel):
        @classmethod
        def from_list(cls, lst: list[str]) -> None:  # type: ignore
            pass

    test_model = TestModel(test="test")  # type: ignore
    assert test_model.test == "test"  # type: ignore


def test_add_custom_field():
    """Test adding a custom field to a class"""

    class TestModel(BaseModel):
        pass

    TestModel.add_custom_field("T", "test", str)
    assert TestModel._get_custom_fields() == [
        Field(line_code="T", attr="test", type=str),
    ]


def test_overwrite_custom_field():
    """Test overwriting a custom field in a class"""

    class TestModel(BaseModel):
        pass

    TestModel.add_custom_field("T", "test", str)
    TestModel.add_custom_field("T", "test2", int)
    assert TestModel._get_custom_fields() == [
        Field(line_code="T", attr="test2", type=int),
    ]


def test_custom_fields_are_not_shared():
    """Test that custom fields are not shared between classes"""

    class TestModel1(BaseModel):
        pass

    class TestModel2(BaseModel):
        pass

    TestModel1.add_custom_field("T", "test", str)
    assert TestModel1._get_custom_fields() == [
        Field(line_code="T", attr="test", type=str),
    ]
    assert not TestModel2._get_custom_fields()


def test_custom_fields_are_inherited():
    """Test that custom fields are inherited by subclasses"""

    class TestModel1(BaseModel):
        pass

    class TestModel2(TestModel1):
        pass

    TestModel1.add_custom_field("T", "test", str)
    assert TestModel1._get_custom_fields() == [
        Field(line_code="T", attr="test", type=str),
    ]
    assert TestModel2._get_custom_fields() == [
        Field(line_code="T", attr="test", type=str),
    ]



from datetime import datetime
from decimal import Decimal

from quiffen.core.category import Category
from quiffen.core.split import Split


def test_create_split():
    """Test creating a split"""
    split = Split()
    assert split.amount is None
    assert split.memo is None
    assert split.category is None

    split2 = Split(amount=Decimal(100), memo="Test Memo")
    assert split2.amount == 100
    assert split2.memo == "Test Memo"
    assert split2.category is None

    split3 = Split(
        amount=Decimal(100),
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    assert split3.amount == 100
    assert split3.memo == "Test Memo"
    assert split3.category is not None
    assert split3.category.name == "Test Category"


def test_eq_success():
    """Test that two splits are equal"""
    split = Split(amount=Decimal(100), memo="Test Memo")
    split2 = Split(amount=Decimal(100), memo="Test Memo")
    assert split == split2


def test_eq_failure():
    """Test that two splits are not equal"""
    split = Split(amount=Decimal(100), memo="Test Memo")
    split2 = Split(amount=Decimal(100), memo="Test Memo2")
    assert split != split2


def test_str_method():
    """Test the string representation of a split"""
    split = Split(amount=Decimal(100), memo="Test Memo")
    assert str(split) == "\n\tSplit:\n\t\tAmount: 100\n\t\tMemo: Test Memo"

    split2 = Split(
        amount=Decimal(100),
        memo="Test Memo",
        category=Category(name="Test Category"),
    )
    assert str(split2) == (
        "\n\tSplit:\n\t\tAmount: 100\n\t\tMemo: Test Memo\n\t\tCategory: Test Category"
    )


def test_to_qif():
    """Test the to_qif method"""
    split = Split(amount=Decimal(100), memo="Test Memo")
    assert split.to_qif() == ("S\n$100\nETest Memo\n")

    test_category = Category(name="Test Category")

    split2 = Split(
        amount=Decimal(100),
        memo="Test Memo",
        category=test_category,
        date=datetime(2019, 1, 1),
        cleared="True",
        check_number=123,
        percent=Decimal(50),
        to_account="Test Account",
        payee_address="Test Address",
    )
    assert split2.to_qif() == (
        "STest Category\n"
        "D2019-01-01\n"
        "$100\n"
        "ETest Memo\n"
        "CTrue\n"
        "L[Test Account]\n"
        "N123\n"
        "%50%\n"
        "ATest Address\n"
    )


def test_to_dict():
    """Test the to_dict method"""
    split = Split(amount=Decimal(100), memo="Test Memo")
    assert split.to_dict() == {
        "amount": 100,
        "memo": "Test Memo",
        "category": None,
        "check_number": None,
        "cleared": None,
        "date": None,
        "percent": None,
        "to_account": None,
        "payee_address": None,
    }

    test_category = Category(name="Test Category")

    split2 = Split(
        amount=Decimal(100),
        memo="Test Memo",
        category=test_category,
    )
    assert split2.to_dict() == {
        "amount": 100,
        "memo": "Test Memo",
        "category": test_category.to_dict(),
        "check_number": None,
        "cleared": None,
        "date": None,
        "percent": None,
        "to_account": None,
        "payee_address": None,
    }


def test_to_dict_with_ignore():
    """Test the to_dict method with ignore"""
    split = Split(amount=Decimal(100), memo="Test Memo")
    assert split.to_dict(ignore={"memo"}) == {
        "amount": 100,
        "category": None,
        "check_number": None,
        "cleared": None,
        "date": None,
        "percent": None,
        "to_account": None,
        "payee_address": None,
    }

    test_category = Category(name="Test Category")

    split2 = Split(
        amount=Decimal(100),
        memo="Test Memo",
        category=test_category,
    )
    assert split2.to_dict(ignore={"memo", "category"}) == {
        "amount": 100,
        "check_number": None,
        "cleared": None,
        "date": None,
        "percent": None,
        "to_account": None,
        "payee_address": None,
    }



from datetime import datetime
from decimal import Decimal

import pytest

from quiffen.core.security import Security


def test_create_security():
    """Test creating a security"""
    security = Security()
    assert security.name is None
    assert security.symbol is None
    assert security.type is None
    assert security.goal is None

    security2 = Security(name="Test Security", symbol="Test Symbol")
    assert security2.name == "Test Security"
    assert security2.symbol == "Test Symbol"
    assert security2.type is None
    assert security2.goal is None

    security3 = Security(
        name="Test Security",
        symbol="Test Symbol",
        type="Test Type",
        goal="Test Goal",
    )
    assert security3.name == "Test Security"
    assert security3.symbol == "Test Symbol"
    assert security3.type == "Test Type"
    assert security3.goal == "Test Goal"


def test_eq_success():
    """Test that two securities are equal"""
    security = Security(name="Test Security", symbol="Test Symbol")
    security2 = Security(name="Test Security", symbol="Test Symbol")
    assert security == security2


def test_eq_failure():
    """Test that two securities are not equal"""
    security = Security(name="Test Security", symbol="Test Symbol")
    security2 = Security(name="Test Security", symbol="Test Symbol2")
    assert security != security2


def test_str_method():
    """Test the string representation of a security"""
    security = Security(name="Test Security", symbol="Test Symbol")
    assert str(security) == ("Security:\n\tName: Test Security\n\tSymbol: Test Symbol")

    security2 = Security(
        name="Test Security",
        symbol="Test Symbol",
        type="Test Type",
        goal="Test Goal",
    )
    assert str(security2) == (
        "Security:\n\t"
        "Name: Test Security\n\t"
        "Symbol: Test Symbol\n\t"
        "Type: Test Type\n\t"
        "Goal: Test Goal"
    )


def test_merge():
    """Test merging two securities"""
    security = Security(name="Test Security", symbol="Test Symbol")
    security2 = Security(
        name="Test Security 2",
        symbol="Test Symbol 2",
        type="Test Type",
        goal="Test Goal",
    )
    security.merge(security2)
    assert security.name == "Test Security"
    assert security.symbol == "Test Symbol"
    assert security.type == "Test Type"
    assert security.goal == "Test Goal"


def test_to_qif():
    """Test the to_qif method"""
    security = Security(name="Test Security", symbol="Test Symbol")
    assert security.to_qif() == ("!Type:Security\nNTest Security\nSTest Symbol\n")

    security2 = Security(
        name="Test Security",
        symbol="Test Symbol",
        type="Test Type",
        goal="Test Goal",
    )
    assert security2.to_qif() == (
        "!Type:Security\nNTest Security\nSTest Symbol\nTTest Type\nGTest Goal\n"
    )


def test_from_list_no_custom_fields():
    """Test creating a security from a list with no custom fields"""
    qif_list = [
        "NTest Security",
        "STest Symbol",
        "TTest Type",
        "GTest Goal",
    ]
    security = Security.from_list(qif_list)
    assert security.name == "Test Security"
    assert security.symbol == "Test Symbol"
    assert security.type == "Test Type"
    assert security.goal == "Test Goal"


def test_from_list_with_custom_fields():
    """Test creating a security from a list with custom fields"""
    setattr(Security, "__CUSTOM_FIELDS", [])  # Reset custom fields
    qif_list = [
        "NTest Security",
        "STest Symbol",
        "TTest Type",
        "GTest Goal",
        "XCustom field 1",
        "Y9238479",
        "DT2022-01-01T00:00:00.000001",
    ]

    # Add custom fields
    Security.add_custom_field(
        line_code="X",
        attr="custom_field_1",
        field_type=str,
    )
    Security.add_custom_field(
        line_code="Y",
        attr="custom_field_2",
        field_type=Decimal,
    )
    Security.add_custom_field(
        line_code="DT",  # Test multi-character line code
        attr="custom_field_3",
        field_type=datetime,
    )

    security = Security.from_list(qif_list)
    assert security.name == "Test Security"
    assert security.symbol == "Test Symbol"
    assert security.type == "Test Type"
    assert security.goal == "Test Goal"
    assert security.custom_field_1 == "Custom field 1"  # type: ignore
    assert security.custom_field_2 == Decimal("9238479")  # type: ignore
    assert security.custom_field_3 == datetime(2022, 1, 1, 0, 0, 0, 1)  # type: ignore
    setattr(Security, "__CUSTOM_FIELDS", [])  # Reset custom fields


def test_from_list_with_unknown_line_code():
    """Test creating a security from a list with an unknown line code"""
    qif_list = [
        "NTest Security",
        "STest Symbol",
        "TTest Type",
        "GTest Goal",
        "ZInvalid field",
    ]

    with pytest.raises(ValueError):
        Security.from_list(qif_list)


def test_from_list_no_symbol():
    """Test creating a security from a list with no custom fields"""
    qif_list = [
        "NTest Security",
        "TTest Type",
        "GTest Goal",
    ]
    security = Security.from_list(qif_list)
    assert security.name == "Test Security"
    # Since we need a symbol, use the name as the symbol
    assert security.symbol == security.name
    assert security.type == "Test Type"
    assert security.goal == "Test Goal"


def test_from_string_default_separator():
    """Test creating a security from a string with the default separator"""
    qif_string = "NTest Security\nSTest Symbol\nTTest Type\nGTest Goal\n"
    security = Security.from_string(qif_string)
    assert security.name == "Test Security"
    assert security.symbol == "Test Symbol"
    assert security.type == "Test Type"
    assert security.goal == "Test Goal"


def test_from_string_custom_separator():
    """Test creating a security from a string with a custom separator"""
    qif_string = "NTest Security---STest Symbol---TTest Type---GTest Goal---"
    security = Security.from_string(qif_string, separator="---")
    assert security.name == "Test Security"
    assert security.symbol == "Test Symbol"
    assert security.type == "Test Type"
    assert security.goal == "Test Goal"


def test_to_dict():
    """Test the to_dict method"""
    security = Security(
        name="Test Security",
        symbol="Test Symbol",
        type="Test Type",
        goal="Test Goal",
    )
    assert security.to_dict() == {
        "name": "Test Security",
        "symbol": "Test Symbol",
        "type": "Test Type",
        "goal": "Test Goal",
        "line_number": None,
    }


def test_to_dict_with_ignore():
    """Test the to_dict method with ignored attributes"""
    security = Security(
        name="Test Security",
        symbol="Test Symbol",
        type="Test Type",
        goal="Test Goal",
    )
    assert security.to_dict(ignore=["type", "line_number"]) == {
        "name": "Test Security",
        "symbol": "Test Symbol",
        "goal": "Test Goal",
    }
